#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
import inspect
import sys
import pickle
import os
import numpy
import pytest
import random
import shutil
import pandas
from cStringIO import StringIO

import bayesdb.utils as utils
from bayesdb.client import Client
from bayesdb.engine import Engine
import bayesdb.bql_grammar as bql

test_tablenames = None
client = None
test_filenames = None

def setup_function(function):
  global test_tablenames, client, test_filenames
  test_tablenames = []
  test_filenames = []
  # Default upgrade_key_column is None, to let the user choose, but need to avoid
  # user input during testing, so for testing just create a new key column.
  client = Client(testing=True)

def teardown_function(function):
  global tablename, client
  for test_tablename in test_tablenames:
    client.engine.drop_btable(test_tablename)

  for test_filename in test_filenames:
    if os.path.exists(test_filename):
        os.remove(test_filename)

def create_dha(path='data/dha.csv', key_column=0):
  test_tablename = 'dhatest' + str(int(time.time() * 1000000)) + str(int(random.random()*10000000))
  csv_file_contents = open(path, 'r').read()
  client('create btable %s from %s' % (test_tablename, path), debug=True, pretty=False, key_column=key_column)

  global test_tablenames
  test_tablenames.append(test_tablename)

  return test_tablename

def test_drop_btable():
  """
  Test to make sure drop btable prompts the user for confirmation, and responds appropriately when
  given certain input.
  """
  import sys
  from cStringIO import StringIO

  # setup the environment
  backup = sys.stdout
  sys.stdout = StringIO()     # capture output

  # TODO: not being tested at all yet...

  out = sys.stdout.getvalue() # release output
  sys.stdout.close()  # close the stream
  sys.stdout = backup # restore original stdout


def test_btable_list():
  global client, test_filenames

  out = set(client('list btables', pretty=False, debug=True)[0]['btable'])
  init_btable_count = len(out)

  test_tablename1 = create_dha()

  out = set(client('list btables', pretty=False, debug=True)[0]['btable'])
  assert len(out) == 1 + init_btable_count
  assert test_tablename1 in out

  test_tablename2 = create_dha()

  out = set(client('list btables', pretty=False, debug=True)[0]['btable'])
  assert len(out) == 2 + init_btable_count
  assert test_tablename1 in out
  assert test_tablename2 in out

  client('drop btable %s' % test_tablename1, yes=True, debug=True, pretty=False)

  out = set(client('list btables', pretty=False, debug=True)[0]['btable'])
  assert len(out) == 1 + init_btable_count
  assert test_tablename1 not in out
  assert test_tablename2 in out

  ## test to make sure btable list is persisted
  del client
  client = Client()

  out = set(client('list btables', pretty=False, debug=True)[0]['btable'])
  assert len(out) == 1 + init_btable_count
  assert test_tablename1 not in out
  assert test_tablename2 in out


def test_save_and_load_models():
  test_tablename1 = create_dha()
  test_tablename2 = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename1), debug=True, pretty=False)
  #client('analyze %s for 1 iteration' % (test_tablename1), debug=True, pretty=False)
  pkl_path = 'test_models.pkl.gz'
  test_filenames.append(pkl_path)
  client('save models from %s to %s' % (test_tablename1, pkl_path), debug=True, pretty=False)
  original_models = client.engine.save_models(test_tablename1)

  client('load models %s into %s' % (pkl_path, test_tablename2), debug=True, pretty=False)
  new_models = client.engine.save_models(test_tablename2)

  assert new_models.values() == original_models.values()

  # Models are saved with schema now, so check that they can be loaded into a table
  # with the same schema that already has models, and can't be loaded into a table
  # with a different schema that already has models.

  # Should work - schemas are the same and table2 already has models
  client('load models %s into %s' % (pkl_path, test_tablename2), debug=True, pretty=False)

  test_tablename3 = create_dha()
  client('update schema for %s set qual_score = categorical' % (test_tablename3), debug=True, pretty=False)

  # Should work - schemas aren't the same, but table3 doesn't have any models, so the schema will be changed.
  client('load models %s into %s' % (pkl_path, test_tablename3), debug=True, pretty=False)

  test_tablename4 = create_dha()
  client('update schema for %s set qual_score = categorical' % (test_tablename4), debug=True, pretty=False)
  client('initialize 2 models for %s' % (test_tablename4), debug=True, pretty=False)

  # Should fail - schemas aren't the same, and table4 already has models, so the models will be incompatible.
  with pytest.raises(utils.BayesDBError):
    client('load models %s into %s' % (pkl_path, test_tablename4), debug=True, pretty=False)

def test_column_lists():
  """ smoke test """
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)

  cname1 = 'cname1'
  cname2 = 'cname2'
  client('show column lists for %s' % test_tablename, debug=True, pretty=False)
  out = client('estimate columns from %s as %s' % (test_tablename, cname1), debug=True, pretty=False)[0]
  assert type(out) == pandas.DataFrame
  assert (out.columns == ['column label', 'column name']).all()

  client('show column lists for %s' % test_tablename, debug=True, pretty=False)
#TODO grammar update, replace tests after implementing show columns for <column_list>
#  client('show columns %s for %s' % (cname1, test_tablename), debug=True, pretty=False)
#  with pytest.raises(utils.BayesDBColumnListDoesNotExistError):
#    client('show columns %s from %s' % (cname2, test_tablename), debug=True, pretty=False)
  out = client('estimate columns from %s order by typicality limit 5 as %s' % (test_tablename, cname1), debug=True, pretty=False)[0]
  assert out.shape == (5, 3)

  client('estimate columns from %s limit 5 as %s' % (test_tablename, cname2), debug=True, pretty=False)
  client('show column lists for %s' % test_tablename, debug=True, pretty=False)
  # TODO same todo as above
  #  client('show columns %s from %s' % (cname1, test_tablename), debug=True, pretty=False)
  #  client('show columns %s from %s' % (cname2, test_tablename), debug=True, pretty=False)

  tmp = 'asdf_test.png'
  test_filenames.append(tmp)
  if os.path.exists(tmp):
    os.remove(tmp)
  # TODO for columns col_name
  client('estimate pairwise dependence probability from %s for %s save to %s' % (test_tablename, cname1, tmp), debug=True, pretty=False)
  test_ast = bql.bql_statement.parseString('estimate pairwise dependence probability from %s for %s save to %s' % (test_tablename, cname1, tmp),parseAll=True)
  assert test_ast.filename == 'asdf_test.png'
  #TODO current parsing breaks save (probably everything) after "for %s"
  #assert os.path.exists(tmp)

  client('estimate pairwise dependence probability from %s for %s' % (test_tablename, cname2), debug=True, pretty=False)

  client('select %s from %s limit 10' % (cname1, test_tablename), debug=True, pretty=False)
  client('select %s from %s limit 10' % (cname2, test_tablename), debug=True, pretty=False)

  client('infer %s from %s with confidence 0.1 limit 10' % (cname1, test_tablename), debug=True, pretty=False)
  client('infer %s from %s with confidence 0.1 limit 10' % (cname2, test_tablename), debug=True, pretty=False)

  client('simulate %s from %s times 10' % (cname1, test_tablename), debug=True, pretty=False)
  client('simulate %s from %s times 10' % (cname2, test_tablename), debug=True, pretty=False)

  # Test dropping column list
  client('drop column list %s from %s' % (cname1, test_tablename), debug=True, pretty=False)
  client('drop column list %s from %s' % (cname2, test_tablename), debug=True, pretty=False)

  # Assert there are now 0 row lists.
  out = client('show column lists for %s' % test_tablename, debug=True, pretty=False)[0]
  assert out.shape == (0, 1)

def test_simulate():
  """ smoke test """
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)
  # TODO given documentation
  assert len(client("simulate qual_score from %s given name='Albany NY' times 5" % test_tablename, debug=True, pretty=False)[0]) == 5
  assert len(client("simulate qual_score from %s given name='Albany NY' and ami_score = 80 times 5" % test_tablename, debug=True, pretty=False)[0]) == 5

  assert len(client("simulate name from %s given name='Albany NY' and ami_score = 80 times 5" % test_tablename, debug=True, pretty=False)[0]) == 5
  assert len(client("simulate name from %s given name='Albany NY', ami_score = 80 times 5" % test_tablename, debug=True, pretty=False)[0]) == 5
  assert len(client("simulate name from %s given name='Albany NY' AND ami_score = 80 times 5" % test_tablename, debug=True, pretty=False)[0]) == 5
  assert len(client("simulate name from %s given ami_score = 80 times 5" % test_tablename, debug=True, pretty=False)[0]) == 5

def test_estimate_columns():
  """ smoke test """
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)

#  client('estimate columns from %s' % test_tablename, debug=True, pretty=False)

  client('estimate columns from %s where typicality > 1' % test_tablename, debug=True, pretty=False)
  client('estimate columns from %s where typicality > 0' % test_tablename, debug=True, pretty=False)
  client('estimate columns from %s where typicality > 0 order by typicality' % test_tablename, debug=True, pretty=False)
#  client('estimate columns from %s order by typicality limit 5' % test_tablename, debug=True, pretty=False)

  client('estimate columns from %s where dependence probability with qual_score > 0' % test_tablename, debug=True, pretty=False)
  client('estimate columns from %s order by dependence probability with qual_score' % test_tablename, debug=True, pretty=False)
  client('estimate columns from %s order by dependence probability with qual_score limit 5' % test_tablename, debug=True, pretty=False)

  out = client('estimate columns from %s order by correlation with qual_score limit 5' % test_tablename, debug=True, pretty=False)[0]
  scores = out["correlation with qual_score"]
  assert (0 <= scores).all() and (scores <= 1).all()
  out = client('estimate columns from %s where correlation with qual_score > 0 order by correlation with qual_score limit 5' % test_tablename, debug=True, pretty=False)[0]
  scores = out["correlation with qual_score"]
  assert (0 <= scores).all() and (scores <= 1).all()

  client('estimate columns from %s order by mutual information with qual_score limit 5' % test_tablename, debug=True, pretty=False)
  client('estimate columns from %s where mutual information with qual_score > 1 order by typicality' % test_tablename, debug=True, pretty=False)

def test_row_clusters():
  """ smoke test """
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)
  row_lists = client('show row lists for %s' % test_tablename, debug=True, pretty=False)[0]
  assert row_lists.shape == (0, 2)
  client('estimate pairwise row similarity from %s save clusters with threshold 0.1 as rcc' % test_tablename, debug=True, pretty=False)
  row_lists = client('show row lists for %s' % test_tablename, debug=True, pretty=False)[0]
  assert row_lists.shape[0] > 0
  client('select * from %s where key in rcc_0' % test_tablename, debug=True, pretty=False)
  #client("select * from %s where similarity to name='McAllen TX' > 0.5 order by similarity to name='McAllen TX' as mcallenrows" % test_tablename, debug=True, pretty=False)
  #client('select * from %s where key in mcallenrows' % test_tablename, debug=True, pretty=False)

  # Test removing row lists
  client('drop row list rcc from %s' % test_tablename, debug=True, pretty=False)
  out = client('show row lists for %s' % test_tablename, debug=True, pretty=False)[0]
  assert out.shape == (0, 2)

def test_select_whereclause_functions():
  """ smoke test """
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)

  # similarity
  client('select name from %s where similarity to 0 > 0' % (test_tablename), debug=True, pretty=False)
  client('select name from %s where similarity to 0 = 0 order by similarity to 0' % (test_tablename), debug=True, pretty=False)
  client('select name from %s where similarity to 1 with respect to qual_score > 0.01' % (test_tablename), debug=True, pretty=False)
  client('select name from %s where similarity to 1 with respect to qual_score, ami_score > 0.01' % (test_tablename), debug=True, pretty=False)

  # row typicality
  client('select * from %s where typicality > 0.04' % (test_tablename), debug=True, pretty=False)
  client('select *, typicality from %s where typicality > 0.06' % (test_tablename), debug=True, pretty=False)

  # predictive probability
  client("select qual_score from %s where predictive probability of qual_score > 0.01" % (test_tablename), debug=True, pretty=False)
  client("select qual_score from %s where predictive probability of name > 0.01" % (test_tablename), debug=True, pretty=False)

  # probability: aggregate, shouldn't work
  with pytest.raises(utils.BayesDBError):
    client('select qual_score from %s where probability of qual_score = 6 > 0.01' % (test_tablename), debug=True, pretty=False)
  with pytest.raises(utils.BayesDBError):
    client("select qual_score from %s where probability of name='Albany NY' > 0.01" % (test_tablename), debug=True, pretty=False)

def test_model_config():
  test_tablename = create_dha()
  global client, test_filenames

  # test naive bayes
  client('initialize 2 models for %s with config naive bayes' % (test_tablename), debug=True, pretty=False)
  #client('analyze %s for 2 iterations wait' % (test_tablename), debug=True, pretty=False)
  client.engine.analyze(test_tablename, model_indices=[0], iterations=2, background=False)
  dep_mat = client('estimate pairwise dependence probability from %s' % test_tablename, debug=True, pretty=False)[0]['matrix']
  ## assert that all dependencies are _0_ (not 1, because there should only be 1 view and 1 cluster!)
  ## except the diagonal, where we've hardcoded every column to be dependent with itself
  assert numpy.all(dep_mat == numpy.identity(dep_mat.shape[0]))

  # test crp
  with pytest.raises(utils.BayesDBNoModelsError):
    client('drop models from %s' % test_tablename, yes=True, debug=True, pretty=False)
  client('initialize 2 models for %s with config crp mixture' % (test_tablename), debug=True, pretty=False)
  #client('analyze %s for 2 iterations wait' % (test_tablename), debug=True, pretty=False)
  client.engine.analyze(test_tablename, model_indices='all', iterations=2, background=False)
  dep_mat = client('estimate pairwise dependence probability from %s' % test_tablename, debug=True, pretty=False)[0]['matrix']
  ## assert that all dependencies are 1 (because there's 1 view, and many clusters)
  ## (with _very_ low probability, this test may fail due to bad luck)
  assert numpy.all(dep_mat == 1)

  # test crosscat
  with pytest.raises(utils.BayesDBNoModelsError):
    client('drop models from %s' % test_tablename, yes=True, debug=True, pretty=False)
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)
  #client('analyze %s for 2 iterations wait' % (test_tablename), debug=True, pretty=False)
  client.engine.analyze(test_tablename, model_indices='all', iterations=2, background=False)
  dep_mat = client('estimate pairwise dependence probability from %s' % test_tablename, debug=True, pretty=False)[0]['matrix']
  ## assert that all dependencies are not all the same
  assert (not numpy.all(dep_mat == 1)) and (not numpy.all(dep_mat == 0))

  # test that you can't change model config
  with pytest.raises(utils.BayesDBError):
    client.engine.initialize_models(test_tablename, 2, 'crp mixture')


@pytest.mark.xfail
def test_analyze():
  """ test designed to make sure that analyze in background runs correct number of iterations """
  # 1 iteration works fine, but multiple iterations don't.
  # AnalyzeMaster is getting called multiple fucking times.
  # 9, 11, 11, 13 analyze_master calls, and 4 iterations done on every model each of those times (except first one did 3 iters once)
  test_tablename = create_dha(key_column=1) # key column is irrelevant, but let's use it
  global client, test_filenames

  models = 3
  out = client('initialize %d models for %s' % (models, test_tablename), debug=True, pretty=False)[0]

  iterations = 3
  out = client('analyze %s for %d iterations' % (test_tablename, iterations), debug=True, pretty=False)[0]

  out = ''
  while 'not currently being analyzed' not in out:
    out = client('show analyze for %s' % test_tablename, debug=True, pretty=False)[0]['message']

  models = client('show models for %s' % test_tablename, debug=True, pretty=False)[0]['models']
  iters_by_model = [v for k,v in models]
  for i in iters_by_model:
    assert i == iterations

def test_using_models():
  """ smoke test """
  test_tablename = create_dha(path='data/dha_missing.csv')
  global client, test_filenames
  client('initialize 3 models for %s' % (test_tablename), debug=True, pretty=False)

  client('select name from %s using model 1' % test_tablename, debug=True, pretty=False)
  with pytest.raises(utils.BayesDBError):
    client('infer name from %s with confidence 0.1 using models 3' % test_tablename, debug=True, pretty=False)
  with pytest.raises(utils.BayesDBError):
    client("simulate qual_score from %s given name='Albany NY' times 5 using models 3" % test_tablename, debug=True, pretty=False)
  with pytest.raises(utils.BayesDBError):
    client('infer name from %s with confidence 0.1 using models 0-3' % test_tablename, debug=True, pretty=False)

  client('infer name from %s with confidence 0.1 limit 10 using models 2' % test_tablename, debug=True, pretty=False)
  client("simulate qual_score from %s given name='Albany NY' times 5 using models 1-2" % test_tablename, debug=True, pretty=False)
  client('estimate columns from %s limit 5 using models 1-2' % test_tablename, debug=True, pretty=False)
  client('estimate pairwise dependence probability from %s using models 1' % (test_tablename), debug=True, pretty=False)
  client('estimate pairwise row similarity from %s save clusters with threshold 0.1 as rcc using models 1-2' % test_tablename, debug=True, pretty=False)

  client('drop model 0 from %s' % test_tablename, debug=True, pretty=False, yes=True)
  with pytest.raises(utils.BayesDBError):
    client('infer name from %s with confidence 0.1 limit 10 using models 0-2' % test_tablename, debug=True, pretty=False)

def test_select():
  """ smoke test """
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)

  client('select name, qual_score from %s' % (test_tablename), debug=True, pretty=False)
  client('select name, qual_score from %s limit 10' % (test_tablename), debug=True, pretty=False)
  client('select name, qual_score from %s order by qual_score limit 10' % (test_tablename), debug=True, pretty=False)
  client('select name, qual_score from %s order by qual_score ASC limit 10' % (test_tablename), debug=True, pretty=False)
  client('select name, qual_score from %s order by qual_score DESC limit 10' % (test_tablename), debug=True, pretty=False)
  client('select * from %s order by qual_score DESC limit 10' % (test_tablename), debug=True, pretty=False)
  client('select name, qual_score from %s where qual_score > 6' % (test_tablename), debug=True, pretty=False)
  client('select * from %s where qual_score > 6' % (test_tablename), debug=True, pretty=False)
  client("select * from %s where qual_score > 80 and name = 'Albany NY'" % (test_tablename), debug=True, pretty=False)
  client("select * from %s where qual_score > 80 and ami_score > 85" % (test_tablename), debug=True, pretty=False)

  # create a column list to be used in future queries
  client('estimate columns from %s limit 5 as clist' % test_tablename, debug=True, pretty=False)

  # similarity
  client('select name, similarity to 0 from %s' % (test_tablename), debug=True, pretty=False)
  client('select name from %s order by similarity to 0' % (test_tablename), debug=True, pretty=False)
  client('select name, similarity to 0 from %s order by similarity to 0' % (test_tablename), debug=True, pretty=False)
  client('select name, similarity to 0 with respect to name from %s order by similarity to 1 with respect to qual_score' % (test_tablename), debug=True, pretty=False)
  client('select name, similarity to 0 from %s order by similarity to 1 with respect to qual_score, ami_score' % (test_tablename), debug=True, pretty=False)
  client('select name, similarity to 0 from %s order by similarity to 1 with respect to clist' % (test_tablename), debug=True, pretty=False)

  # Add some cases to be sure that referencing row index and column values produces the same output.
  out1 = client('select name, qual_score, similarity to 161 from %s order by similarity to 161 limit 5' % (test_tablename), debug=True, pretty=False)[0]
  out2 = client('select name, qual_score, similarity to name = "McAllen TX" from %s order by similarity to name = "McAllen TX" limit 5' % (test_tablename), debug=True, pretty=False)[0]
  out3 = client('select name, qual_score, similarity to key = 161 from %s order by similarity to key = 161 limit 5' % (test_tablename), debug=True, pretty=False)[0]

  # Assert columns are equal
  for col_idx in range(out1.shape[1]):
    assert (out1.iloc[col_idx] == out2.iloc[col_idx]).all()
    assert (out2.iloc[col_idx] == out3.iloc[col_idx]).all()

  # row typicality
  client('select typicality from %s' % (test_tablename), debug=True, pretty=False)
  client('select *, typicality from %s' % (test_tablename), debug=True, pretty=False)
  client('select typicality from %s order by typicality limit 10' % (test_tablename), debug=True, pretty=False)

  # probability
  # why is this so slow, when predictive probability is really fast? these are _observed_
  # for qual_score (numerical): probability takes 20 times longer than predictive prob (about 5 seconds total for 300 rows)
  # for name (categorical): probability takes extremely long (about 75 seconds for 300 rows)
  #  while predictive probability takes under one second for 300 rows
  st = time.time()
  client('select probability of qual_score = 6 from %s' % (test_tablename), debug=True, pretty=False)
  el = time.time() - st
  st = time.time()
  client("select probability of name='Albany NY' from %s" % (test_tablename), debug=True, pretty=False)
  el2 = time.time() - st

  #client("select name from %s order by probability of name='Albany NY' DESC" % (test_tablename), debug=True, pretty=False)
  # TODO: test that probability function doesn't get evaluated 2x for each row
  #client("select probability of name='Albany NY' from %s order by probability of name='Albany NY' DESC" % (test_tablename), debug=True, pretty=False)

  # predictive probability
  # these are really fast! :) simple predictive probability, unobserved
  client("select predictive probability of qual_score from %s" % (test_tablename), debug=True, pretty=False)
  client("select predictive probability of name from %s" % (test_tablename), debug=True, pretty=False)
  client("select predictive probability of qual_score from %s order by predictive probability of name" % (test_tablename), debug=True, pretty=False)
  client("select predictive probability of qual_score from %s order by predictive probability of qual_score" % (test_tablename), debug=True, pretty=False)

  ## Aggregate functions: can't order by these.

  # mutual information
  client("select name, qual_score, mutual information of name with qual_score from %s" % (test_tablename), debug=True, pretty=False)

  # dependence probability
  client("select dependence probability of name with qual_score from %s" % (test_tablename), debug=True, pretty=False)
  client("select name, qual_score, dependence probability of name with qual_score from %s" % (test_tablename), debug=True, pretty=False)

  # correlation
  client("select name, qual_score, correlation of name with qual_score from %s" % (test_tablename), debug=True, pretty=False)

  # column typicality
  client("select typicality of qual_score, typicality of name from %s" % (test_tablename), debug=True, pretty=False)
  client("select typicality of qual_score from %s" % (test_tablename), debug=True, pretty=False)

  # correlation with missing values
  test_tablename = create_dha(path='data/dha_missing.csv')
  client("select name, qual_score, correlation of name with qual_score from %s" % (test_tablename), debug=True, pretty=False)

def test_into():
  test_tablename = create_dha()
  global client

  client('drop btable test_btable_select', yes=True)

  # Test that select can produce a new btable with INTO, and that it can be analyzed and manipulated like other btables
  client('select name, qual_score from %s limit 5 into test_btable_select' % test_tablename, debug=True, pretty=False)
  out = client('select * from test_btable_select', debug=True, pretty=False)[0]
  assert len(out) == 5
  assert (out.columns == ['key', 'name', 'qual_score']).all()

  client('summarize select * from test_btable_select')
  client('label columns for test_btable_select set qual_score = quality')

  client('initialize 2 models for test_btable_select')
  client('analyze test_btable_select for 2 iterations')
  client('simulate * from test_btable_select times 5')

  client('drop btable test_btable_select', yes=True)

def test_pandas():
  test_tablename = create_dha()
  global client

  # Test that output is a dict if pretty=False and pandas_output=False
  out = client("select name, qual_score from %s limit 10" % (test_tablename), debug=True, pretty=False, pandas_output=False)
  assert type(out[0]) == dict

  # Test that output is pandas DataFrame when pretty=False and a table-like object is returned (pandas_output=True by default)
  out = client("select name, qual_score from %s limit 10" % (test_tablename), debug=True, pretty=False)
  assert type(out[0]) == pandas.DataFrame

  # Test that it still works when no rows are returned
  client("select name, qual_score from %s where qual_score < 0" % (test_tablename), debug=True, pretty=False)

  # Get the returned data frame from the first list element of the previous result.
  test_df = out[0]

  # Test creation of a btable from pandas DataFrame
  client("drop btable %s" % (test_tablename), yes=True)
  client("create btable %s from pandas" % (test_tablename), debug=True, pretty=False, pandas_df=test_df, key_column=1)

def test_summarize():
  test_tablename = create_dha()
  global client

  # Test that the output is a pandas DataFrame when pretty=False
  out = client('summarize select name, qual_score from %s' % (test_tablename), debug=True, pretty=False)[0]
  assert type(out) == pandas.DataFrame
  assert (out.columns == ['', 'name', 'qual_score']).all()

  # Test that stats from summary_describe and summary_freqs made it into the output DataFrame
  # Note that all of these stats won't be present in EVERY summarize output, but all should be in the output
  # from the previous test.
  expected_indices = ['type', 'count', 'unique', 'mean', 'std', 'min', '25%', '50%', '75%', 'max', \
    'mode', 'prob_mode']
  assert all([x in list(out['']) for x in expected_indices])

  # Test that it works on columns of predictive functions.
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)
  client('summarize select correlation of name with qual_score from %s' % (test_tablename), debug=True, pretty=False)

  # Test with fewer than 5 unique values (output should have fewer rows)
  out = client('summarize select name, qual_score from %s limit 3' % (test_tablename), debug=True, pretty=False)[0]
  assert out.shape == (12, 3)

  # Test with no rows
  out = client('summarize select name, qual_score from %s where qual_score < 0' % (test_tablename), debug=True, pretty=False)[0]
  assert out.shape == (0, 3)

  # Test with only a discrete column
  client('summarize select name from %s' % (test_tablename), debug=True, pretty=False)

  # Test with only a numerical column
  client('summarize select qual_score from %s' % (test_tablename), debug=True, pretty=False)

def test_select_where_col_equal_val():
  test_tablename = create_dha()
  global client, test_filenames
  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)
  basic_similarity = client('select * from %s where similarity to 1 > .6 limit 5' % (test_tablename),pretty=False, debug=True)[0]['key']
  col_val_similarity = client('select * from %s where similarity to name = "Akron OH" > .6 limit 5' % (test_tablename),pretty=False, debug=True)[0]['key']
  assert len(basic_similarity) == len(col_val_similarity)

def test_labeling():
  test_tablename = create_dha()
  global client, test_filenames

  client('label columns for %s set name = Name of the hospital, qual_score = Overall quality score' % (test_tablename), debug=True, pretty=False)
  client('show label for %s name, qual_score' % (test_tablename), debug=True, pretty=False)
  client('show label for %s' % (test_tablename), debug=True, pretty=False)

  # Test getting columns from CSV
  client('label columns for %s from data/dha_labels.csv' % (test_tablename), debug=True, pretty=False)

def test_user_metadata():
  test_tablename = create_dha()
  global client, test_filenames

  client('update metadata for %s set data_source = Dartmouth Atlas of Health, url = http://www.dartmouthatlas.org/tools/downloads.aspx' % (test_tablename), debug=True, pretty=False)
  client('update metadata for %s from data/dha_user_metadata.csv' % (test_tablename), debug=True, pretty=False)

  client('show metadata for %s data_source, url' % (test_tablename), debug=True, pretty=False)

  # Test that show metadata also works when no keys are specified
  client('show metadata for %s' % (test_tablename), debug=True, pretty=False)

def test_freq_hist():
  test_tablename = create_dha()
  global client, test_filenames

  # Test that freq and hist work and return a DataFrame
  out = client('freq select qual_score from %s' % (test_tablename), debug=True, pretty=False)[0]
  assert type(out) == pandas.DataFrame
  assert out['qual_score'][0] == 87.5
  assert out['frequency'][0] == 7

  out = client('hist select qual_score from %s' % (test_tablename), debug=True, pretty=False)[0]
  assert type(out) == pandas.DataFrame
  assert out.shape == (10, 4)
  assert out['frequency'][0] == 1

  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)
  client.engine.analyze(tablename=test_tablename, iterations=2, background=False)

  # Results for infer should match select, since there are no missing values
  out = client('freq infer qual_score from %s with confidence 0' % (test_tablename), debug=True, pretty=False)[0]
  assert type(out) == pandas.DataFrame
  assert out['qual_score'][0] == 87.5
  assert out['frequency'][0] == 7

  out = client('hist infer qual_score from %s with confidence 0' % (test_tablename), debug=True, pretty=False)[0]
  assert type(out) == pandas.DataFrame
  assert out.shape == (10, 4)
  assert out['frequency'][0] == 1

  # For simulate, we just have to go by size and expected range
  out = client('freq simulate qual_score from %s times 20' % (test_tablename), debug=True, pretty=False)[0]
  assert out.shape[1] == 3
  assert (out['probability'] < 1).all()

  out = client('hist simulate qual_score from %s times 20' % (test_tablename), debug=True, pretty=False)[0]
  assert out.shape[1] == 4
  assert (out['frequency'] <= 20).all()
  assert (out['probability'] < 1).all()

def test_update_schema():
  test_tablename = create_dha()
  global client, test_filenames

  # Test setting one column to each type other than numerical
  out = client('update schema for %s set qual_score = ignore, ami_score = categorical, pneum_score = cyclic(0, 100)' % (test_tablename), debug=True, pretty=False)[0]
  assert (out['datatype'][out['column'] == 'qual_score'] == 'ignore').all()
  assert (out['datatype'][out['column'] == 'ami_score'] == 'categorical').all()
  assert (out['datatype'][out['column'] == 'pneum_score'] == 'cyclic').all()

  # Test setting categorical with a cardinality parameter
  out = client('update schema for %s set name = categorical(350)' % (test_tablename), debug=True, pretty=False)[0]
  assert (out['datatype'][out['column'] == 'name'] == 'categorical').all()

  # Selecting qual_score should still work even after it's ignored, also should work in where clauses and order by clauses
  client('select qual_score from %s' % (test_tablename), debug=True, pretty=False)
  out = client('select name, qual_score, ami_score from %s where qual_score > 90 order by qual_score' % (test_tablename), debug=True, pretty=False)[0]
  assert (out['qual_score'] > 90).all()
  assert (out['qual_score'] == out['qual_score'].order(ascending=False)).all()

  # Also test where clause with ignored text column
  client('update schema for %s set name = ignore' % (test_tablename), debug=True, pretty=False)
  out = client('select name, qual_score from %s where name = "Albany NY"' % (test_tablename), debug=True, pretty=False)[0]
  assert out.shape == (1, 3)
  assert (out['name'] == "Albany NY").all()

  # Set qual_score, ami_score, pneum_score, and total_fte back to numerical, and select should work again
  client('update schema for %s set qual_score = numerical, ami_score = numerical, pneum_score = numerical, total_fte = numerical' % (test_tablename), debug=True, pretty=False)

  # Set back to ignore, run models, and then estimation shouldn't work for qual_score
  client('update schema for %s set qual_score = ignore' % (test_tablename), debug=True, pretty=False)

  client('initialize 2 models for %s' % (test_tablename), debug=True, pretty=False)

  client.engine.analyze(tablename=test_tablename, iterations=2, background=False)

  with pytest.raises(utils.BayesDBError):
    # Next two statements should fail because they attempt functions on an 'ignore' column
    client('estimate columns from %s order by correlation with qual_score limit 5' % (test_tablename), debug=True, pretty=False)
    client('estimate columns from %s order by dependence probability with qual_score limit 5' % (test_tablename), debug=True, pretty=False)
    # Next two statements should fail because they 1) try to set a new key and 2) try to change the key's type
    client('update schema for %s set name = key' % (test_tablename), debug=True, pretty=False)
    client('update schema for %s set key = numerical' % (test_tablename), debug=True, pretty=False)
    # Next two statements should fail because they set parameters that don't contain the data.
    client('update schema for %s set name = categorical(3)' % (test_tablename), debug=True, pretty=False)
    client('update schema for %s set qual_score = cyclic(0, 10)' % (test_tablename), debug=True, pretty=False)



import os


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


# make sure cwd is correct
this_file = os.path.abspath(__file__)
this_dir = os.path.split(this_file)[0]
os.chdir(this_dir)


setup(
    name='BayesDB',
    version='0.2.0',
    author='MIT Probabilistic Computing Project',
    author_email = 'bayesdb@mit.edu',
    url='probcomp.csail.mit.edu/bayesdb',
    long_description='BayesDB',
    packages=['bayesdb', 'bayesdb.tests'],
    install_requires=['jsonrpc', 'requests', 'Twisted', 'pyOpenSSL',
                      'numpy', 'scipy', 'matplotlib>=1.2.0',
                      'Sphinx', 'pytest',
                      'prettytable', 'cmd2', 'pyparsing>=2.0.1',
                      'ipython', 'pandas>=0.13',
                      ],
    license='Apache License, Version 2.0',
    entry_points = """
                   [console_scripts]
                   bql = bayesdb.bql:run_command_line
                   """
                   
)



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Dan Lovell and Jay Baxter
#   Authors: Dan Lovell, Baxter Eaves, Jay Baxter, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import requests
import json


global_id = 0

def create_message(method_name, params, id):
    id += 1
    message = {
        'jsonrpc': '2.0',
        'method': method_name,
        'params': params,
        'id': str(id),
        }
    json_message = json.dumps(message)
    return json_message, id

def call(method_name, args_dict, URI, id=None, print_message=False):
    global global_id
    if id is None: id = global_id
    message, id = create_message(method_name, args_dict, id)
    global_id = global_id + 1
    if print_message: print 'trying message:', message
    r = requests.put(URI, data=message)
    r.raise_for_status()
    out = json.loads(r.content)
    #
    if isinstance(out, dict) and 'result' in out:
        out = out['result']
    else:
        print "call(%s, <args_dict>, %s): ERROR" % (method_name, URI)
    return out, id

def call_and_print(method_name, args_dict, URI, id=0):
    out, id = call(method_name, args_dict, URI, id=id, print_message=True)
    print out
    print
    return out, id



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import utils
import os
import bql_grammar as bql
import pyparsing as pp
import functions
import operator


class Parser(object):
    def __init__(self):
        self.reset_root_dir()

    def pyparse_input(self, input_string):
        """
        Uses the grammar defined in bql_grammar to create a pyparsing object out of an input string
        :param str input_string: the input block of bql statement(s)
        :return: the pyparsing object for the parsed statement.
        :rtype: pyparsing.ParseResults
        :raises BayesDBParseError: if the string is not a valid bql statement.
        """
        try:
            bql_blob_ast = bql.bql_input.parseString(input_string, parseAll=True)
        except pp.ParseException as x:
            raise utils.BayesDBParseError("Invalid query. Could not parse (Line {e.lineno}, column {e.col}):\n\t'{e.line}'\n\t".format(e=x) + ' ' * x.col + '^')
        return bql_blob_ast

    def parse_single_statement(self,bql_statement_ast):
        """
        Calls parse_method on the statement_id in bql_statement_ast.
        :param bql_statement_ast pyparsing.ParseResults: The single statement pyparsing result
        :return: parse_method(bql_statement_ast)
        """
        ## TODO Check for nest
        parse_method = getattr(self,'parse_' + bql_statement_ast.statement_id)
        return parse_method(bql_statement_ast)

#####################################################################################
## -------------------------- Individual Parse Methods --------------------------- ##
#####################################################################################

    def parse_list_btables(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('list_btables', args_dict, None):
        """
        if bql_statement_ast.statement_id == "list_btables":
            return 'list_btables', dict(), None
        else:
            raise utils.BayesDBParseError("Parsing statement as LIST BTABLES failed")

    def parse_help(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('help', args_dict, None):
        """
        method= None
        if bql_statement_ast.method_name != '':
            method = bql_statement_ast.method_name
        return 'help', dict(method=method), None

    def parse_execute_file(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('execute_file', args_dict, None):
        """
        return 'execute_file', dict(filename=self.get_absolute_path(bql_statement_ast.filename)), None

    def parse_show_schema(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_schema', args_dict, None):
        """
        return 'show_schema', dict(tablename=bql_statement_ast.btable), None

    def parse_show_models(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_models', args_dict, None):
        """
        return 'show_models', dict(tablename=bql_statement_ast.btable), None

    def parse_show_diagnostics(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_diagnostics', args_dict, None):
        """
        return 'show_diagnostics', dict(tablename=bql_statement_ast.btable), None

    def parse_drop_column_list(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('drop_column_list', args_dict, None):
        """
        list_name = None
        if bql_statement_ast.list_name != '':
            list_name = bql_statement_ast.list_name
        return 'drop_column_list', dict(tablename=bql_statement_ast.btable, list_name=list_name), None

    def parse_drop_row_list(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('drop_row_list', args_dict, None):
        """
        list_name = None
        if bql_statement_ast.list_name != '':
            list_name = bql_statement_ast.list_name
        return 'drop_row_list', dict(tablename=bql_statement_ast.btable, list_name=list_name), None

    def parse_drop_models(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('drop_models', args_dict, None):
        """
        model_indices = None
        if bql_statement_ast.index_clause != '':
            model_indices = bql_statement_ast.index_clause.asList()
        return 'drop_models', dict(tablename=bql_statement_ast.btable, model_indices=model_indices), None

    def parse_initialize_models(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('initialize_models', args_dict, None):
        """
        n_models = 16 ##TODO magic number - move to config file
        if bql_statement_ast.num_models != '':
            n_models = int(bql_statement_ast.num_models)
        tablename = bql_statement_ast.btable
        arguments_dict = dict(tablename=tablename, n_models=n_models, model_config=None)
        if bql_statement_ast.config != '':
            arguments_dict['model_config'] = bql_statement_ast.config
        return 'initialize_models', arguments_dict, None

    def parse_create_btable(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('create_btable', args_dict, client_dict):
        """
        tablename = bql_statement_ast.btable
        filename = self.get_absolute_path(bql_statement_ast.filename)
        client_dict = dict(csv_path = filename)
        if bql_statement_ast.codebook_file != '':
            codebook_path = self.get_absolute_path(bql_statement_ast.codebook_file)
            client_dict['codebook_path'] = codebook_path
        return 'create_btable', dict(tablename=tablename, cctypes_full=None), client_dict
        #TODO types?

    def parse_upgrade_btable(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('upgrade_btable', args_dict, client_dict):
        """
        tablename = bql_statement_ast.btable
        return 'upgrade_btable', dict(tablename=tablename), None

    def parse_describe(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('describe', args_dict, None):
        """
        tablename = None
        if bql_statement_ast.btable != '':
            tablename = bql_statement_ast.btable
        columnset = None
        if bql_statement_ast.columnset != '':
            columnset = bql_statement_ast.columnset

        return 'describe', dict(tablename=tablename, columnset=columnset), None

    def parse_update_codebook(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('update_codebook', args_dict, client_dict):
        """
        tablename = None
        if bql_statement_ast.btable != '':
            tablename = bql_statement_ast.btable

        codebook_path = None
        if bql_statement_ast.filename != '':
            codebook_path = bql_statement_ast.filename

        return 'update_codebook', dict(tablename=tablename), dict(codebook_path=codebook_path)

    def parse_update_short_names(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('update_short_names', args_dict, None):
        """
        tablename = None
        if bql_statement_ast.btable != '':
            tablename = bql_statement_ast.btable

        mappings = None
        if bql_statement_ast.label_clause != '':
            mappings = {}
            for label_set in bql_statement_ast.label_clause:
                mappings[label_set[0]] = label_set[1]

        return 'update_short_names', dict(tablename=tablename, mappings=mappings), None

    def parse_update_descriptions(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('update_desriptions', args_dict, None):
        """
        tablename = None
        if bql_statement_ast.btable != '':
            tablename = bql_statement_ast.btable

        mappings = None
        if bql_statement_ast.label_clause != '':
            mappings = {}
            for label_set in bql_statement_ast.label_clause:
                mappings[label_set[0]] = label_set[1]

        return 'update_descriptions', dict(tablename=tablename, mappings=mappings), None

    def parse_update_schema(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('update_schema', args_dict, None):
        """
        tablename = bql_statement_ast.btable
        mappings = dict()
        type_clause = bql_statement_ast.type_clause
        for update in type_clause:
            column = update[0]
            cctype = update[1]
            mappings[column] = dict()
            mappings[column]['cctype'] = cctype
            if 'parameters' in update:
                mappings[column]['parameters'] = dict()
                for param_type in update.parameters.keys():
                    mappings[column]['parameters'][param_type] = update.parameters[param_type]
            else:
                mappings[column]['parameters'] = None
        return 'update_schema', dict(tablename=tablename, mappings=mappings), None

    def parse_drop_btable(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('drop_btable', args_dict, client_dict):
        """
        return 'drop_btable', dict(tablename=bql_statement_ast.btable), None

    def parse_analyze(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('analyze', args_dict, None):
        """
        model_indices = None
        iterations = None
        seconds = None
        kernel = 0
        wait = True
        tablename = bql_statement_ast.btable
        if bql_statement_ast.index_clause != '':
            model_indices = bql_statement_ast.index_clause.asList()
        if bql_statement_ast.num_minutes !='':
            seconds = int(bql_statement_ast.num_minutes) * 60
        if bql_statement_ast.num_iterations !='':
            iterations = int(bql_statement_ast.num_iterations)
        if bql_statement_ast.with_kernel_clause != '':
            kernel = bql_statement_ast.with_kernel_clause.kernel_id
            if kernel == 'mh': ## TODO should return None or something for invalid kernels
                kernel=1
        if bql_statement_ast.wait != '':
            wait = False
        return 'analyze', dict(tablename=tablename, model_indices=model_indices,
                               iterations=iterations, seconds=seconds,
                               ct_kernel=kernel, background=wait), None

    def parse_cancel_analyze(self, bql_statement_ast): ##TODO no tests written
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('cancel_analyze', args_dict, None):
        """
        return 'cancel_analyze', dict(tablename=bql_statement_ast.btable), None

    def parse_show_analyze(self, bql_statement_ast): ##TODO no tests written
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_analyze', args_dict, None):
        """
        return 'show_analyze', dict(tablename=bql_statement_ast.btable), None

    def parse_show_row_lists(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_row_lists', args_dict, None):
        """
        return 'show_row_lists', dict(tablename=bql_statement_ast.btable), None

    def parse_show_column_lists(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_column_lists', args_dict, None):
        """
        return 'show_column_lists', dict(tablename=bql_statement_ast.btable), None

    def parse_show_columns(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_columns', args_dict, None):
        """
        return 'show_columns', dict(tablename=bql_statement_ast.btable, column_list=bql_statement_ast.column_list[0]), None

    def parse_save_models(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('save_models', args_dict, client_dict):
        """
        return 'save_models', dict(tablename=bql_statement_ast.btable), dict(pkl_path=bql_statement_ast.filename)

    def parse_load_models(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('load_models', args_dict, client_dict):
        """
        return 'load_models', dict(tablename=bql_statement_ast.btable), dict(pkl_path=bql_statement_ast.filename)

    def parse_label_columns(self, bql_statement_ast): ##TODO only smoke test right now
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('label_columns', args_dict, client_dict):
        """
        tablename = bql_statement_ast.btable
        source = None
        mappings = None
        if bql_statement_ast.label_clause != '':
            source = 'inline'
            mappings = {}
            for label_set in bql_statement_ast.label_clause:
                mappings[label_set[0]] = label_set[1]
        csv_path = None
        if bql_statement_ast.filename != '':
            csv_path = bql_statement_ast.filename
            source = 'file'
        return 'label_columns', \
            dict(tablename=tablename, mappings=mappings), \
            dict(source=source, csv_path=csv_path)

    def parse_show_metadata(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_metadata', args_dict, None):
        """
        tablename = None
        if bql_statement_ast.btable != '':
            tablename = bql_statement_ast.btable
        keyset = None
        if bql_statement_ast.keyset != '':
            keyset = bql_statement_ast.keyset
        return 'show_metadata', dict(tablename=tablename, keyset=keyset), None

    def parse_show_label(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('show_label', args_dict, None):
        """
        tablename = None
        if bql_statement_ast.btable != '':
            tablename = bql_statement_ast.btable
        columnset = None
        if bql_statement_ast.columnset != '':
            columnset = bql_statement_ast.columnset
        return 'show_labels', dict(tablename=tablename, columnset=columnset), None

    def parse_update_metadata(self, bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('update_metadata', args_dict, client_dict):
        """
        tablename = bql_statement_ast.btable
        source = None
        mappings = None
        if bql_statement_ast.label_clause != '':
            source = 'inline'
            mappings = {}
            for label_set in bql_statement_ast.label_clause:
                mappings[label_set[0]] = label_set[1]
        csv_path = None
        if bql_statement_ast.filename != '':
            csv_path = bql_statement_ast.filename
            source = 'file'
        return 'update_metadata', \
            dict(tablename=tablename, mappings=mappings), \
            dict(source=source, csv_path=csv_path)

    def parse_query(self, bql_statement_ast):
        '''
        master parser for queries (select, infer, simulate, estimate pairwise, etc)
        returns a general args dict which the specific versions of those functions
        will then trim and check for illegal aruguments through assertions.

        :param bql_statement_ast pyparsing.ParseResults:
        :return (statement_id, args_dict, client_dict):
        '''
        statement_id = bql_statement_ast.statement_id

        confidence = None
        if bql_statement_ast.confidence != '':
            confidence = float(bql_statement_ast.confidence)
        filename = None
        if bql_statement_ast.filename != '':
            filename = bql_statement_ast.filename
        functions = bql_statement_ast.functions
        givens = None
        if bql_statement_ast.given_clause != '':
            givens = bql_statement_ast.given_clause
        limit = float('inf')
        if bql_statement_ast.limit != '':
            limit = int(bql_statement_ast.limit)
        modelids = None
        if bql_statement_ast.using_models_index_clause != '':
            modelids = bql_statement_ast.using_models_index_clause.asList()
        name = None
        if bql_statement_ast.as_column_list != '':
            ## TODO name is a bad name
            name = bql_statement_ast.as_column_list
        newtablename = None
        if bql_statement_ast.newtablename != '':
            newtablename = bql_statement_ast.newtablename
        numpredictions = None
        if bql_statement_ast.times != '':
            numpredictions = int(bql_statement_ast.times)
        numsamples = None
        if bql_statement_ast.samples != '':
            numsamples = int(bql_statement_ast.samples)
        order_by = False
        if bql_statement_ast.order_by != '':
            order_by = bql_statement_ast.order_by
        plot=(bql_statement_ast.plot == 'plot')
        column_list = None
        row_list = None
        if bql_statement_ast.for_list != '':
            if statement_id == 'estimate_pairwise':
                ##TODO can only handle single column or row lists in this clause.
                column_list = bql_statement_ast.for_list.asList()
                assert len(column_list) == 1, "BayesDBParseError: You may only specify one column list. Remove %s" % column_list[1]
                column_list = column_list[0]
            elif statement_id == 'estimate_pairwise_row':
                row_list = bql_statement_ast.for_list.asList()
                assert len(row_list) == 1, "BayesDBParseError: You may only specify one row list. Remove %s" % row_list[1]
                row_list = row_list[0]

            else:
                raise utils.BayesDBParseError("Invalid query: FOR <list> only acceptable in estimate pairwise.")

        summarize=(bql_statement_ast.summarize == 'summarize')
        hist = (bql_statement_ast.hist == 'hist')
        freq = (bql_statement_ast.freq == 'freq')
        tablename = bql_statement_ast.btable
        clusters_name = None
        threshold = None
        if bql_statement_ast.clusters_clause != '':
            clusters_name = bql_statement_ast.clusters_clause.as_label
            threshold = float(bql_statement_ast.clusters_clause.threshold)
        whereclause = None
        if bql_statement_ast.where_conditions != '':
            whereclause = bql_statement_ast.where_conditions
        return statement_id, \
            dict(clusters_name=clusters_name,
                 confidence = confidence,
                 functions=functions,
                 givens=givens,
                 limit=limit,
                 modelids=modelids,
                 name=name,
                 newtablename=newtablename,
                 numpredictions=numpredictions,
                 numsamples=numsamples,
                 order_by=order_by,
                 plot=plot,
                 column_list=column_list,
                 row_list=row_list,
                 summarize=summarize,
                 hist=hist,
                 freq=freq,
                 tablename=tablename,
                 threshold=threshold,
                 whereclause=whereclause), \
            dict(plot=plot,
                 scatter=False, ##TODO remove scatter from args
                 pairwise=False, ##TODO remove pairwise from args
                 filename=filename)

    def parse_infer(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('infer', args_dict, client_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        tablename = args_dict['tablename']
        functions = args_dict['functions']
        summarize = args_dict['summarize']
        hist = args_dict['hist']
        freq = args_dict['freq']
        plot = args_dict['plot']
        whereclause = args_dict['whereclause']
        limit = args_dict['limit']
        order_by = args_dict['order_by']
        modelids = args_dict['modelids']
        newtablename = args_dict['newtablename']
        confidence = args_dict['confidence']
        numsamples = args_dict['numsamples']

        pairwise = client_dict['pairwise']
        filename = client_dict['filename']
        scatter = client_dict['scatter']

        assert args_dict['clusters_name'] == None, "BayesDBParsingError: SAVE CLUSTERS clause not allowed in INFER"
        assert args_dict['threshold'] == None, "BayesDBParsingError: SAVE CLUSTERS clause not allowed in INFER"
        assert args_dict['givens'] == None, "BayesDBParsingError: GIVENS clause not allowed in INFER"
        assert args_dict['name'] == None, "BayesDBParsingError: SAVE AS <column_list> clause not allowed in INFER"
        assert args_dict['numpredictions'] == None, "BayesDBParsingError: TIMES clause not allowed in INFER"
        assert args_dict['column_list'] == None, "BayesDBParsingError: FOR <columns> clause not allowed in INFER"
        assert args_dict['row_list'] == None, "BayesDBParsingError: FOR <rows> not allowed in INFER"
        for function in functions:
            assert function.function_id == '', "BayesDBParsingError: %s not valid in INFER" % function.function_id

        assert confidence is None or 0.0 <= confidence <= 1.0, "BayesDBParsingError: CONFIDENCE must be unspecified (default 0) or in the interval [0, 1]"

        return 'infer', \
            dict(tablename=tablename, functions=functions,
                 newtablename=newtablename, confidence=confidence,
                 whereclause=whereclause, limit=limit,
                 numsamples=numsamples, order_by=order_by,
                 plot=plot, modelids=modelids, summarize=summarize, hist=hist, freq=freq), \
            dict(plot=plot, scatter=scatter, pairwise=pairwise, filename=filename)

    def parse_select(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('select', args_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        tablename = args_dict['tablename']
        functions = args_dict['functions']
        summarize = args_dict['summarize']
        hist = args_dict['hist']
        freq = args_dict['freq']
        plot = args_dict['plot']
        whereclause = args_dict['whereclause']
        limit = args_dict['limit']
        order_by = args_dict['order_by']
        modelids = args_dict['modelids']
        newtablename = args_dict['newtablename']
        numsamples = args_dict['numsamples']

        pairwise = client_dict['pairwise']
        filename = client_dict['filename']
        scatter = client_dict['scatter']

        assert args_dict['clusters_name'] == None, "BayesDBParsingError: SAVE CLUSTERS clause not allowed in SELECT"
        assert args_dict['threshold'] == None, "BayesDBParsingError: SAVE CLUSTERS clause not allowed in SELECT"
        assert args_dict['givens'] == None, "BayesDBParsingError: GIVENS clause not allowed in SELECT"
        assert args_dict['name'] == None, "BayesDBParsingError: SAVE AS <column_list> clause not allowed in SELECT"
        assert args_dict['numpredictions'] == None, "BayesDBParsingError: TIMES clause not allowed in SELECT"
        assert args_dict['column_list'] == None, "BayesDBParsingError: FOR <columns> clause not allowed in SELECT"
        assert args_dict['row_list'] == None, "BayesDBParsingError: FOR <rows> not allowed in SELECT"
        assert args_dict['confidence'] == None, "BayesDBParsingError: CONFIDENCE not allowed in SELECT"

        for function in functions:
            assert function.conf == '', "BayesDBParsingError: CONF (WITH CONFIDENCE) not valid in SELECT"
        return 'select', \
            dict(tablename=tablename, whereclause=whereclause,
                 functions=functions, limit=limit, order_by=order_by, plot=plot, numsamples=numsamples,
                 modelids=modelids, summarize=summarize, hist=hist, freq=freq, newtablename=newtablename), \
            dict(pairwise=pairwise, scatter=scatter, filename=filename, plot=plot)

    def parse_simulate(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('simulate', args_dict, client_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        tablename = args_dict['tablename']
        functions = args_dict['functions']
        summarize = args_dict['summarize']
        hist = args_dict['hist']
        freq = args_dict['freq']
        plot = args_dict['plot']
        order_by = args_dict['order_by']
        modelids = args_dict['modelids']
        newtablename = args_dict['newtablename']
        givens = args_dict['givens']
        numpredictions = args_dict['numpredictions']

        pairwise = client_dict['pairwise']
        filename = client_dict['filename']
        scatter = client_dict['scatter']

        # Require the number of observations to simulate
        assert args_dict['numpredictions'] is not None, "BayesDBParsingError: Include TIMES: SIMULATE <columns> FROM <btable> TIMES <times>."

        assert args_dict['clusters_name'] == None, "BayesDBParsingError: SAVE CLUSTERS clause not allowed in SIMULATE."
        assert args_dict['numsamples'] == None, 'BayesDBParsingError: WITH <numsamples> SAMPLES clause not allowed in SIMULATE.'
        assert args_dict['threshold'] == None, "BayesDBParsingError: SAVE CLUSTERS clause not allowed in SIMULATE."
        assert args_dict['name'] == None, "BayesDBParsingError: SAVE AS <column_list> clause not allowed in SIMULATE."
        assert args_dict['column_list'] == None, "BayesDBParsingError: FOR <columns> clause not allowed in SIMULATE."
        assert args_dict['row_list'] == None, "BayesDBParsingError: FOR <rows> not allowed in SIMULATE."
        assert args_dict['confidence'] == None, "BayesDBParsingError: CONFIDENCE not allowed in SIMULATE."
        assert args_dict['whereclause'] == None, "BayesDBParsingError: whereclause not allowed in SIMULATE. Use GIVEN instead."
        for function in functions:
            assert function.function_id == '', "BayesDBParsingError: %s not valid in SIMULATE" % function.function_id
            assert function.conf == '', "BayesDBParsingError: CONF (WITH CONFIDENCE) not valid in SIMULATE"

        return 'simulate', \
            dict(tablename=tablename, functions=functions,
                 newtablename=newtablename, givens=givens,
                 numpredictions=numpredictions, order_by=order_by,
                 plot=plot, modelids=modelids, summarize=summarize, hist=hist, freq=freq), \
            dict(filename=filename, plot=plot, scatter=scatter, pairwise=pairwise)

    def parse_estimate(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('estimate', args_dict, client_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        assert args_dict['functions'][0] == 'column', "BayesDBParseError: must be ESTIMATE COLUMNS."
        functions = args_dict['functions']
        tablename = args_dict['tablename']
        whereclause = args_dict['whereclause']
        limit = args_dict['limit']
        order_by = args_dict['order_by']
        modelids = args_dict['modelids']
        name = args_dict['name']
        numsamples = args_dict['numsamples']

        assert args_dict['clusters_name'] == None, "BayesDBParsingError: SAVE CLUSTERS not allowed in estimate columns."
        assert args_dict['confidence'] == None, "BayesDBParsingError: WITH CONFIDENCE not allowed in estimate columns."
        assert args_dict['givens'] == None, "BayesDBParsingError: GIVENS not allowed in estimate columns."
        assert args_dict['newtablename'] == None, "BayesDBParsingError: INTO TABLE not allowed in estimate columns."
        assert args_dict['numpredictions'] == None, "BayesDBParsingError: TIMES not allowed in estimate columns."
        assert args_dict['column_list'] == None, "BayesDBParsingError: FOR COLUMNS not allowed in estimate columns."
        assert args_dict['row_list'] == None, "BayesDBParsingError: FOR ROWS not allowed in estimate columns."
        assert args_dict['summarize'] == False, "BayesDBParsingError: SUMMARIZE not allowed in estimate columns."
        assert args_dict['hist'] == False, "BayesDBParsingError: HIST not allowed in estimated columns."
        assert args_dict['freq'] == False, "BayesDBParsingError: FREQ not allowed in estimated columns."
        assert args_dict['threshold'] == None, "BayesDBParsingError: SAVE CLUSTERS not allowed in estimate columns."
        assert args_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in estimate columns."

        assert client_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in estimate columns."
        assert client_dict['scatter'] == False, "BayesDBParsingError: SCATTER not allowed in estimate columns."
        assert client_dict['pairwise'] == False, "BayesDBParsingError: PAIRWISE not allowed in estimate columns."
        assert client_dict['filename'] == None, "BayesDBParsingError: AS FILE not allowed in estimate columns."

        return 'estimate_columns', \
            dict(tablename=tablename, functions=functions,
                 whereclause=whereclause, limit=limit, numsamples=numsamples,
                 order_by=order_by, name=name, modelids=modelids), \
            None

    def parse_create_column_list(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('create_column_list', args_dict, client_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        functions = args_dict['functions']
        name = args_dict['name']
        tablename = args_dict['tablename']

        assert args_dict['clusters_name'] == None, "BayesDBParsingError: SAVE CLUSTERS not allowed in create column list."
        assert args_dict['confidence'] == None, "BayesDBParsingError: WITH CONFIDENCE not allowed in create column list."
        assert args_dict['givens'] == None, "BayesDBParsingError: GIVENS not allowed in create column list."
        assert args_dict['newtablename'] == None, "BayesDBParsingError: INTO TABLE not allowed in create column list."
        assert args_dict['numpredictions'] == None, "BayesDBParsingError: TIMES not allowed in create column list."
        assert args_dict['column_list'] == None, "BayesDBParsingError: FOR COLUMNS not allowed in create column list."
        assert args_dict['row_list'] == None, "BayesDBParsingError: FOR ROWS not allowed in create column list."
        assert args_dict['summarize'] == False, "BayesDBParsingError: SUMMARIZE not allowed in create column list."
        assert args_dict['hist'] == False, "BayesDBParsingError: HIST not allowed in estimated columns."
        assert args_dict['freq'] == False, "BayesDBParsingError: FREQ not allowed in estimated columns."
        assert args_dict['threshold'] == None, "BayesDBParsingError: SAVE CLUSTERS not allowed in create column list."
        assert args_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in create column list."
        assert args_dict['numsamples'] == None, "BayesDBParsingError: NUMSAMPLES not allowed in create column list."
        assert args_dict['modelids'] == None, "BayesDBParsingError: USING MODELS not allowed in create column list."
        assert args_dict['order_by'] == False, "BayesDBParsingError: ORDER BY not allowed in create column list."
        assert args_dict['limit'] == float('inf'), "BayesDBParsingError: LIMIT not allowed in create column list."
        assert args_dict['whereclause'] == None, "BayesDBParsingError: WHERE not allowed in create column list."

        assert client_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in create column list."
        assert client_dict['scatter'] == False, "BayesDBParsingError: SCATTER not allowed in create column list."
        assert client_dict['pairwise'] == False, "BayesDBParsingError: PAIRWISE not allowed in create column list."
        assert client_dict['filename'] == None, "BayesDBParsingError: AS FILE not allowed in create column list."

        return 'create_column_list', \
            dict(tablename=tablename, functions=functions, name=name), None


    def parse_estimate_pairwise_row(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('estimate_pairwise_row', args_dict, client_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        functions = args_dict['functions'][0]
        assert len(args_dict['functions']) == 1, "BayesDBParsingError: Only one function allowed in estimate pairwise."
        tablename = args_dict['tablename']
        row_list = args_dict['row_list']
        clusters_name = args_dict['clusters_name']
        threshold = args_dict['threshold']
        modelids = args_dict['modelids']
        filename = client_dict['filename']

        assert args_dict['confidence'] == None, "BayesDBParsingError: WITH CONFIDENCE not allowed in ESTIMATE PAIRWISE."
        assert args_dict['givens'] == None, "BayesDBParsingError: GIVENS not allowed in ESTIMATE PAIRWISE."
        assert args_dict['newtablename'] == None, "BayesDBParsingError: INTO TABLE not allowed in ESTIMATE PAIRWISE."
        assert args_dict['numpredictions'] == None, "BayesDBParsingError: TIMES not allowed in ESTIMATE PAIRWISE."
        assert args_dict['numsamples'] == None, "BayesDBParsingError: WITH SAMPLES not allowed in ESTIMATE PAIRWISE."
        assert args_dict['column_list'] == None, "BayesDBParsingError: FOR COLUMNS not allowed in ESTIMATE PAIRWISE."
        assert args_dict['summarize'] == False, "BayesDBParsingError: SUMMARIZE not allowed in ESTIMATE PAIRWISE."
        assert args_dict['hist'] == False, "BayesDBParsingError: HIST not allowed in ESTIMATE PAIRWISE."
        assert args_dict['freq'] == False, "BayesDBParsingError: FREQ not allowed in ESTIMATE PAIRWISE."
        assert args_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in ESTIMATE PAIRWISE."
        assert args_dict['whereclause'] == None, "BayesDBParsingError: WHERE not allowed in ESTIMATE PAIRWISE."

        assert client_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in ESTIMATE PAIRWISE."
        assert client_dict['scatter'] == False, "BayesDBParsingError: SCATTER not allowed in ESTIMATE PAIRWISE."
        assert client_dict['pairwise'] == False, "BayesDBParsingError: PAIRWISE not allowed in ESTIMATE PAIRWISE."

        return 'estimate_pairwise_row', \
            dict(tablename=tablename, function=functions,
                 row_list=row_list, clusters_name=clusters_name,
                 threshold=threshold, modelids=modelids), \
            dict(filename=filename)

    def parse_estimate_pairwise(self,bql_statement_ast):
        """
        :param bql_statement_ast pyparsing.ParseResults:
        :return ('estimate_pairwise', args_dict, client_dict):
        """
        method_name, args_dict, client_dict = self.parse_query(bql_statement_ast)
        functions = args_dict['functions']
        assert len(args_dict['functions']) == 1, "BayesDBParsingError: Only one function allowed in estimate pairwise."
        assert functions[0].function_id in ['correlation', 'mutual information', 'dependence probability']
        function_name = functions[0].function_id

        tablename = args_dict['tablename']
        column_list = args_dict['column_list']
        clusters_name = args_dict['clusters_name']
        threshold = args_dict['threshold']
        modelids = args_dict['modelids']
        numsamples = args_dict['numsamples']

        filename = client_dict['filename']


        assert args_dict['confidence'] == None, "BayesDBParsingError: WITH CONFIDENCE not allowed in ESTIMATE PAIRWISE."
        assert args_dict['givens'] == None, "BayesDBParsingError: GIVENS not allowed in ESTIMATE PAIRWISE."
        assert args_dict['newtablename'] == None, "BayesDBParsingError: INTO TABLE not allowed in ESTIMATE PAIRWISE."
        assert args_dict['numpredictions'] == None, "BayesDBParsingError: TIMES not allowed in ESTIMATE PAIRWISE."
        assert args_dict['row_list'] == None, "BayesDBParsingError: FOR ROWS not allowed in ESTIMATE PAIRWISE."
        assert args_dict['summarize'] == False, "BayesDBParsingError: SUMMARIZE not allowed in ESTIMATE PAIRWISE."
        assert args_dict['hist'] == False, "BayesDBParsingError: HIST not allowed in ESTIMATE PAIRWISE."
        assert args_dict['freq'] == False, "BayesDBParsingError: FREQ not allowed in ESTIMATE PAIRWISE."
        assert args_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in ESTIMATE PAIRWISE."
        assert args_dict['whereclause'] == None, "BayesDBParsingError: whereclause not allowed in ESTIMATE PAIRWISE"

        assert client_dict['plot'] == False, "BayesDBParsingError: PLOT not allowed in ESTIMATE PAIRWISE."
        assert client_dict['scatter'] == False, "BayesDBParsingError: SCATTER not allowed in ESTIMATE PAIRWISE."
        assert client_dict['pairwise'] == False, "BayesDBParsingError: PAIRWISE not allowed in ESTIMATE PAIRWISE."

        return 'estimate_pairwise', \
            dict(tablename=tablename, function_name=function_name, numsamples=numsamples,
                 column_list=column_list, clusters_name=clusters_name,
                 threshold=threshold, modelids=modelids), \
            dict(filename=filename)

#####################################################################################
## ------------------------------ Function parsing ------------------------------- ##
#####################################################################################
    def get_args_pred_prob(self, function_group, M_c):
        """
        :param function_group pyparsing.ParseResults:
        :param M_c:
        :return c_idx: column index
        :rtype int:
        returns the column index from a predictive probability function
        raises exceptions for unfound columns
        """
        if function_group.column != '' and function_group.column in M_c['name_to_idx']:
            column = function_group.column
            c_idx = M_c['name_to_idx'][column]
            return c_idx
        elif function_group.column != '':
            raise utils.BayesDBParseError("Invalid query: could not parse '%s'" % function_group.column)
        else:
            raise utils.BayesDBParseError("Invalid query: missing column argument")

    def get_args_prob(self,function_group, M_c):
        """
        :param function_group:
        Returns column_index, value from a probability function
        raises exception for unfound columns
        """
        if function_group.column != '' and function_group.column in M_c['name_to_idx']:
            column = function_group.column
            c_idx = M_c['name_to_idx'][column]
        elif function_group.column != '':
            raise utils.BayesDBParseError("Invalid query: could not parse '%s'" % function_group.column)
        else:
            raise utils.BayesDBParseError("Invalid query: missing column argument")
        value = utils.string_to_column_type(function_group.value, column, M_c)
        return c_idx, value

    def get_args_similarity(self,function_group, M_c, M_c_full, T, T_full, column_lists):
        """
        returns the target_row_id and a list of with_respect_to columns based on
        similarity function
        Raises exception for unfound columns
        """
        ##TODO some cleaining with row_clause
        target_row_id = None
        target_columns = None
        if function_group != '':
            ## Case for given row_id
            if function_group.row_id != '':
                target_row_id = int(function_group.row_id)
            ## Case for format column = value
            elif function_group.column != '':
                assert T is not None
                assert M_c_full is not None
                target_col_name = function_group.column
                target_col_value = function_group.column_value
                target_row_id = utils.row_id_from_col_value(target_col_value, target_col_name, M_c_full, T_full)
        ## With respect to clause
        with_respect_to_clause = function_group.with_respect_to
        if with_respect_to_clause !='':

            column_set = with_respect_to_clause.column_list
            target_column_names = []
            for column_name in column_set:
                if column_name == '*':
                    target_columns = None
                    break
                elif column_lists is not None and column_name in column_lists.keys():
                    target_column_names += column_lists[column_name]
                elif column_name in M_c['name_to_idx']:
                    target_column_names.append(column_name)
                else:
                    raise utils.BayesDBParseError("Invalid query: column '%s' not found" % column_name)
            target_columns = [M_c['name_to_idx'][column_name] for column_name in target_column_names]
        return target_row_id, target_columns

    def get_args_typicality(self,function_group, M_c):
        """
        returns column_index if present, if not, returns True. ##TODO this needs a ton of testing
        if invalid column, raises exception
        """
        if function_group.column == '':
            return True
        else:
            return utils.get_index_from_colname(M_c, function_group.column)

    def get_args_pairwise_column(self,function_group, M_c):
        """
        designed to handle dependence probability, mutual information, and correlation function_groups
        all have an optional of clause
        returns of_column_index, with_column_index
        invalid column raises exception
        """
        with_column = function_group.with_column
        with_column_index = utils.get_index_from_colname(M_c, with_column)
        of_column_index = None
        if function_group.of_column != '':
            of_column_index = utils.get_index_from_colname(M_c, function_group.of_column)
        return of_column_index, with_column_index

#####################################################################################
## ----------------------------- Sub query parsing  ------------------------------ ##
########################parse_where#############################################################

    def parse_where_clause(self, where_clause_ast, M_c, T, M_c_full, T_full, column_lists):
        """
        Creates conditions: the list of conditions in the whereclause
        List of (c_idx, op, val)
        """
        conditions = []
        operator_map = {'<=': operator.le, '<': operator.lt, '=': operator.eq,
                        '>': operator.gt, '>=': operator.ge, 'in': operator.contains}

        for single_condition in where_clause_ast:
            ## Confidence in the where clause not yet handled by client/engine.
            confidence = None
            if single_condition.conf != '':
                confidence = float(single_condition.conf)

            raw_value = single_condition.value
            function = None
            args = None
            ## SELECT and INFER versions
            if single_condition.function.function_id == 'typicality':
                value = utils.value_string_to_num(raw_value)
                function = functions._row_typicality
                assert self.get_args_typicality(single_condition.function, M_c) == True
                args = True
            elif single_condition.function.function_id == 'similarity':
                value = utils.value_string_to_num(raw_value)
                function = functions._similarity
                args = self.get_args_similarity(single_condition.function, M_c, M_c_full, T, T_full, column_lists)
            elif single_condition.function.function_id == 'predictive probability':
                value = utils.value_string_to_num(raw_value)
                function = functions._predictive_probability
                args = self.get_args_pred_prob(single_condition.function, M_c)

            elif single_condition.function.function_id == 'key':
                value = raw_value
                function = functions._row_id
            elif single_condition.function.column != '':
                ## whereclause of the form "where col = val"
                column_name = single_condition.function.column
                assert column_name != '*'
                if column_name in M_c['name_to_idx']:
                    args = (M_c['name_to_idx'][column_name], confidence)
                    value = utils.string_to_column_type(raw_value, column_name, M_c)
                    function = functions._column
                elif column_name in M_c_full['name_to_idx']:
                    args = M_c_full['name_to_idx'][column_name]
                    value = raw_value
                    function = functions._column_ignore
                else:
                    raise utils.BayesDBParseError("Invalid where clause: column %s was not found in the table" %
                                                  column_name)
            else:
                if single_condition.function.function_id != '':
                    raise utils.BayesDBParseError("Invalid where clause: %s not allowed." %
                                                  single_condition.function.function_id)
                else:
                    raise utils.BayesDBParseError("Invalid where clause. Unrecognized function")
            if single_condition.operation != '':
                op = operator_map[single_condition.operation]
            else:
                raise utils.BayesDBParseError("Invalid where clause: no operator found")
            conditions.append((function, args, op, value))
        return conditions

    def parse_column_whereclause(self, whereclause, M_c, T): ##TODO throw exception on parseable, invalid
        """
        Creates conditions: the list of conditions in the whereclause
        List of (c_idx, op, val)
        """
        conditions = []
        if whereclause == None:
            return conditions
        operator_map = {'<=': operator.le, '<': operator.lt, '=': operator.eq,
                        '>': operator.gt, '>=': operator.ge, 'in': operator.contains}

        for single_condition in whereclause:
            ## Confidence in the where clause not yet handled by client/engine.

            raw_value = single_condition.value
            value = utils.value_string_to_num(raw_value)
            function = None
            args = None
            _ = None
            ## SELECT and INFER versions
            if single_condition.function.function_id == 'typicality':
                function = functions._col_typicality
                assert self.get_args_typicality(single_condition.function, M_c) == True
                args = None
            elif single_condition.function.function_id == 'dependence probability':
                function = functions._dependence_probability
                _, args = self.get_args_pairwise_column(single_condition.function, M_c)
            elif single_condition.function.function_id == 'mutual information':
                function = functions._mutual_information
                _, args = self.get_args_pairwise_column(single_condition.function, M_c)
            elif single_condition.function.function_id == 'correlation':
                function = functions._correlation
                _, args = self.get_args_pairwise_column(single_condition.function, M_c)
            else:
                if single_condition.function.function_id != '':
                    raise utils.BayesDBParseError("Invalid where clause: %s not allowed." %
                                                  single_condition.function.function_id)
                else:
                    raise utils.BayesDBParseError("Invalid where clause. Unrecognized function")
            if single_condition.operation != '':
                op = operator_map[single_condition.operation]
            else:
                raise utils.BayesDBParseError("Invalid where clause: no operator found")
            if _ != None:
                raise utils.BayesDBParseError("Invalid where clause, do not specify an 'of' column in estimate columns")
            conditions.append(((function, args), op, value))
        return conditions

    def parse_order_by_clause(self, order_by_clause_ast, M_c, T, M_c_full, T_full, column_lists):
        function_list = []
        for orderable in order_by_clause_ast:
            confidence = None
            if orderable.conf != '':
                confidence = float(orderable.conf)
            desc = True
            if orderable.asc_desc == 'asc':
                desc = False
            if orderable.function.function_id == 'similarity':
                function = functions._similarity
                args = self.get_args_similarity(orderable.function, M_c, M_c_full, T, T_full, column_lists)
            elif orderable.function.function_id == 'typicality':
                function = functions._row_typicality
                args = self.get_args_typicality(orderable.function, M_c)
            elif orderable.function.function_id == 'predictive probability':
                function = functions._predictive_probability
                args = self.get_args_pred_prob(orderable.function, M_c)
            elif orderable.function.column != '':
                if orderable.function.column in M_c['name_to_idx']:
                    function = functions._column
                    args = (M_c['name_to_idx'][orderable.function.column], confidence)
                else:
                    function = functions._column_ignore
                    args = M_c_full['name_to_idx'][orderable.function.column]
            else:
                raise utils.BayesDBParseError("Invalid order by clause.")
            function_list.append((function, args, desc))
        return function_list

    def parse_column_order_by_clause(self, order_by_clause_ast, M_c, ):
        function_list = []
        for orderable in order_by_clause_ast:
            desc = True
            if orderable.asc_desc == 'asc':
                desc = False
            if orderable.function.function_id == 'typicality':
                assert orderable.function.column == '', "BayesDBParseError: Column order by typicality cannot include 'of %s'" % orderable.function.column
                function = functions._col_typicality
                args = None
            elif orderable.function.function_id == 'dependence probability':
                function = functions._dependence_probability
                _, args = self.get_args_pairwise_column(orderable.function, M_c)
            elif orderable.function.function_id == 'correlation':
                function = functions._correlation
                _, args = self.get_args_pairwise_column(orderable.function, M_c)
            elif orderable.function.function_id == 'mutual information':
                function = functions._mutual_information
                _, args = self.get_args_pairwise_column(orderable.function, M_c)
            else:
                raise utils.BayesDBParseError("Invalid order by clause. Can only order by typicality, correlation, mutual information, or dependence probability.")
            function_list.append((function, args, desc))

        return function_list

    def parse_functions(self, function_groups, M_c=None, T=None, M_c_full=None, T_full=None, column_lists=None, key_column_name=None):
        '''
        Generates two lists of functions, arguments, aggregate tuples.
        Returns queries, query_colnames

        queries is a list of (query_function, query_args, aggregate) tuples,
        where query_function is: row_id, column, probability, similarity.

        For row_id: query_args is ignored (so it is None).
        For column: query_args is a c_idx.
        For probability: query_args is a (c_idx, value) tuple.
        For similarity: query_args is a (target_row_id, target_column) tuple.
        '''
        ## Return the table key as the first column - should only be None if called from simulate
        if key_column_name is not None:
            query_colnames = [key_column_name]
            index_list, name_list, ignore_column = self.parse_column_set(key_column_name, M_c, M_c_full, column_lists)
            queries = [(functions._column_ignore, column_index, False) for column_index in index_list]
        else:
            query_colnames = []
            queries = []

        for function_group in function_groups:
            if function_group.function_id == 'predictive probability':
                queries.append((functions._predictive_probability,
                                self.get_args_pred_prob(function_group, M_c),
                                False))
                query_colnames.append(' '.join(function_group))
            elif function_group.function_id == 'typicality':
                if function_group.column != '':
                    queries.append((functions._col_typicality,
                                    self.get_args_typicality(function_group, M_c),
                                    True))
                else:
                    queries.append((functions._row_typicality,
                                    self.get_args_typicality(function_group, M_c),
                                    False))
                query_colnames.append(' '.join(function_group))
            elif function_group.function_id == 'probability':
                queries.append((functions._probability,
                                self.get_args_prob(function_group, M_c),
                                True))
                query_colnames.append(' '.join(function_group))
            elif function_group.function_id == 'similarity':
                assert M_c is not None
                assert M_c_full is not None
                queries.append((functions._similarity,
                                self.get_args_similarity(function_group, M_c, M_c_full, T, T_full, column_lists),
                                False))
                pre_name_list = function_group.asList()
                if function_group.with_respect_to != '':
                    pre_name_list[-1] = ', '.join(pre_name_list[-1])
                query_colnames.append(' '.join(pre_name_list))
            elif function_group.function_id == 'dependence probability':
                queries.append((functions._dependence_probability,
                                self.get_args_pairwise_column(function_group, M_c),
                                True))
                query_colnames.append(' '.join(function_group))
            elif function_group.function_id == 'mutual information':
                queries.append((functions._mutual_information,
                                self.get_args_pairwise_column(function_group, M_c),
                                True))
                query_colnames.append(' '.join(function_group))
            elif function_group.function_id == 'correlation':
                queries.append((functions._correlation,
                                self.get_args_pairwise_column(function_group, M_c),
                                True))
                query_colnames.append(' '.join(function_group))
            ## single column, column_list, or *
            elif function_group.column_id not in ['', key_column_name]:
                column_name = function_group.column_id
                confidence = None
                assert M_c is not None
                index_list, name_list, ignore_column = self.parse_column_set(column_name, M_c, M_c_full, column_lists)
                if ignore_column:
                    queries += [(functions._column_ignore, column_index, False) for column_index in index_list]
                else:
                    if function_group.conf != '':
                        confidence = float(function_group.conf)
                    queries += [(functions._column, (column_index, confidence), False) for column_index in index_list]
                if confidence is not None:
                    query_colnames += [name + ' with confidence %s' % confidence for name in name_list]
                else:
                    query_colnames += [name for name in name_list]
            elif function_group.column_id == key_column_name:
                # Key column will be added to the output anyways.
                pass
            else:
                raise utils.BayesDBParseError("Invalid query: could not parse function")
        return queries, query_colnames

    def parse_column_set(self, column_name, M_c, M_c_full, column_lists = None):
        """
        given a string representation of a column name or column_list,
        returns a list of the column indexes, list of column names.
        """
        index_list = []
        name_list = []
        ignore_column = False
        if column_name == '*':
            all_columns = utils.get_all_column_names_in_original_order(M_c)
            index_list += [M_c['name_to_idx'][column_name] for column_name in all_columns]
            name_list += [name for name in all_columns]
        elif (column_lists is not None) and (column_name in column_lists.keys()):
            index_list += [M_c['name_to_idx'][name] for name in column_lists[column_name]]
            name_list += [name for name in column_lists[column_name]]
        elif column_name in M_c['name_to_idx']:
            index_list += [M_c['name_to_idx'][column_name]]
            name_list += [column_name]
        elif column_name in M_c_full['name_to_idx']:
            index_list += [M_c_full['name_to_idx'][column_name]]
            name_list += [column_name]
            ignore_column = True
        else:
            raise utils.BayesDBParseError("Invalid query: %s not found." % column_name)
        return index_list, name_list, ignore_column

#####################################################################################
## --------------------------- Other Helper functions ---------------------------- ##
#####################################################################################

    def set_root_dir(self, root_dir):
        """Set the root_directory, used as the base for all relative paths."""
        self.root_directory = root_dir

    def reset_root_dir(self):
        """Set the root_directory, used as the base for all relative paths, to
        the current working directory."""
        self.root_directory = os.getcwd()

    def get_absolute_path(self, relative_path):
        """
        If a relative file path is given by the user in a command,
        this method is used to convert the path to an absolute path
        by assuming that the correct base directory is self.root_directory.
        """
        relative_path = os.path.expanduser(relative_path)
        if os.path.isabs(relative_path):
            return relative_path
        else:
            return os.path.join(self.root_directory, relative_path)



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
import sys
import datetime
import json
import cPickle as pickle
#import pickle
import shutil
import contextlib
import threading
import re

from threading import RLock

import data_utils
import utils

import bayesdb.settings as S


class ModelLocks():
    """
    Creates individual model-level locks.
    Any user method always needs to acquire locks in ascending order to avoid deadlock!!!!
    """

    def __init__(self, persistence_layer):
        self.tablename_dict = dict()
        self.persistence_layer = persistence_layer

        for tablename in self.persistence_layer.btable_index:
            self.tablename_dict[tablename] = dict()
            modelids = self.persistence_layer.get_model_ids(tablename)
            for modelid in modelids:
                self.tablename_dict[tablename][modelid] = RLock()

    def add_tablename_if_not_exist(self, tablename):
        if tablename not in self.tablename_dict:
            self.tablename_dict[tablename] = dict()

    def add_model_if_not_exist(self, tablename, modelid):
        if modelid not in self.tablename_dict[tablename]:
            self.tablename_dict[tablename][modelid] = RLock()

    def acquire(self, tablename, modelid):
        self.add_tablename_if_not_exist(tablename)
        self.add_model_if_not_exist(tablename, modelid)

        self.tablename_dict[tablename][modelid].acquire()

    def release(self, tablename, modelid):
        if modelid in self.tablename_dict[tablename]:
            self.tablename_dict[tablename][modelid].release()

    def acquire_table(self, tablename):
        self.add_tablename_if_not_exist(tablename)
        for modelid in sorted(self.tablename_dict[tablename].keys()):
            self.tablename_dict[tablename][modelid].acquire()

    def release_table(self, tablename):
        for modelid, lock in self.tablename_dict[tablename].items():
            # Only release locks this thread owns. There could be a case where
            # a new model was created while he had the table lock.
            if threading.current_thread().ident == lock._RLock__owner:
                lock.release()

    def drop(self, tablename, modelid):
        del self.tablename_dict[tablename][modelid]

    def drop_all(self, tablename):
        self.tablename_dict[tablename] = dict()


class PersistenceLayer(object):
    """
    Stores btables in the following format in the "data" directory:
    bayesdb/data/
    ..btable_index.pkl
    ..<tablename>/
    ....metadata_full.pkl
    ....metadata.pkl
    ....column_labels.pkl
    ....column_lists.pkl
    ....row_lists.pkl
    ....models/
    ......diagnostic_data.pkl
    ......model_<id>.pkl

    table_index.pkl: list of btable names.

    metadata.pkl: dict. keys: M_r, M_c, T, cctypes
    metadata_full.pkl: dict. keys: M_r_full, M_c_full, T_full, cctypes_full
    column_lists.pkl: dict. keys: column list names, values: list of column names.
    row_lists.pkl: dict. keys: row list names, values: list of row keys (need to update all these if
        table key is changed!).
    models.pkl: dict[model_idx] -> dict[X_L, X_D, iterations, column_crp_alpha, logscore, num_views,
        model_config]. Idx starting at 1.
    diagnostic_data.pkl: dict[model_idx] -> dict[score, num_views, min_cluster_view,
        mean_cluster_view, max_cluster_view, ...more to come] each entry of which is dict with keys
        'value' and 'step'. For example, the 'score' diagnostic for model 9 which has been
        storing diagnostic data every 5 transition steps for 10 transitions may look like:
        diagnostic_data[9]['score'] => {'value': [-1023.0246, -842.8801, -820.3913],
        'step':[0, 4, 9]}
    data.csv: the raw csv file that the data was loaded from.

    Should be a Singleton class, but Python doesn't enforce that. Could be refactored as just a
    module.
    """

    def __init__(self):
        """
        Create data directory if doesn't exist: every other function requires data_dir.
        """
        self.cur_dir = os.path.dirname(os.path.abspath(__file__))
        self.data_dir = os.path.join(self.cur_dir, 'data')
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)
        self.load_btable_index()  # sets self.btable_index
        self.model_locks = ModelLocks(self)
        self.btable_check_index = list()

    def load_btable_index(self):
        """
        Create btable_index.pkl with an empty list if it doesn't exist; otherwise, read its
        contents.
        Set it to self.btable_index
        """
        btable_index_path = os.path.join(self.data_dir, 'btable_index.pkl')
        if not os.path.exists(btable_index_path):
            self.btable_index = []
            self.write_btable_index()
        else:
            f = open(btable_index_path, 'rb')
            self.btable_index = pickle.load(f)
            f.close()

    def write_btable_index(self):
        btable_index_path = os.path.join(self.data_dir, 'btable_index.pkl')
        f = open(btable_index_path, 'wb')
        pickle.dump(self.btable_index, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    def add_btable_to_index(self, tablename):
        self.btable_index.append(tablename)
        self.write_btable_index()

    def remove_btable_from_index(self, tablename):
        self.btable_index.remove(tablename)
        self.write_btable_index()

    def get_metadata(self, tablename):
        try:
            x = os.path.join(self.data_dir, tablename, 'metadata.pkl')
            f = open(os.path.join(self.data_dir, tablename, 'metadata.pkl'), 'rb')
        except Exception as e:
            raise utils.BayesDBError("Error: metadata does not exist. Has %s been corrupted?"
                                     % self.data_dir)
        try:
            metadata = pickle.load(f)
        except Exception as e:
            raise utils.BayesDBError("Error: metadata file could not be loaded for table %s"
                                     % tablename)
        f.close()
        return metadata

    def get_metadata_full(self, tablename):
        try:
            f = open(os.path.join(self.data_dir, tablename, 'metadata_full.pkl'), 'rb')
        except Exception as e:
            raise utils.BayesDBError("Error: metadata_full file doesn't exist. This is most "
                                     "likely a result of this btable being created with an old "
                                     "version of BayesDB. Please try recreating the table from "
                                     "the original csv, and loading any models you might have.")
        metadata = pickle.load(f)
        f.close()
        return metadata

    def write_metadata(self, tablename, metadata):
        metadata_f = open(os.path.join(self.data_dir, tablename, 'metadata.pkl'), 'wb')
        pickle.dump(metadata, metadata_f, pickle.HIGHEST_PROTOCOL)
        metadata_f.close()

    def write_metadata_full(self, tablename, metadata):
        metadata_f = open(os.path.join(self.data_dir, tablename, 'metadata_full.pkl'), 'wb')
        pickle.dump(metadata, metadata_f, pickle.HIGHEST_PROTOCOL)
        metadata_f.close()

    def drop_column_list(self, tablename, list_name):
        column_lists = self.get_column_lists(tablename)
        if list_name in column_lists:
            column_lists.pop(list_name, None)
        elif '%s_0' % list_name in column_lists:
            suffix = 0
            while '%s_%d' % (list_name, suffix) in column_lists:
                column_lists.pop('%s_%d' % (list_name, suffix), None)
                suffix += 1
        else:
            raise utils.BayesDBColumnListDoesNotExistError(list_name, tablename)
        self.write_column_lists(tablename, column_lists)

    def drop_row_list(self, tablename, list_name):
        row_lists = self.get_row_lists(tablename)
        if list_name in row_lists:
            row_lists.pop(list_name, None)
        elif '%s_0' % list_name in row_lists:
            suffix = 0
            while '%s_%d' % (list_name, suffix) in row_lists:
                row_lists.pop('%s_%d' % (list_name, suffix), None)
                suffix += 1
        else:
            raise utils.BayesDBRowListDoesNotExistError(list_name, tablename)
        self.write_row_lists(tablename, row_lists)

    def get_model_config(self, tablename):
        """
        Just loads one model, and gets the model_config from it.
        """
        model = self.get_models(tablename, modelid=self.get_max_model_id(tablename))
        if model is None:
            return None
        if 'model_config' in model:
            return model['model_config']
        else:
            return None

    def get_models(self, tablename, modelid=None):
        """
        Return the models dict for the table if modelid is None.
        If modelid is an int, then return the model specified by that id.
        If modelid is a list, then get each individual model specified by each int in that list.
        """
        models_dir = os.path.join(self.data_dir, tablename, 'models')
        if os.path.exists(models_dir):
            if modelid is not None:
                def get_single_model(modelid):
                    self.model_locks.acquire(tablename, modelid)
                    # Only return one of the models
                    full_fname = os.path.join(models_dir, 'model_%d.pkl' % modelid)
                    if not os.path.exists(full_fname):
                        self.model_locks.release(tablename, modelid)
                        return None
                    f = open(full_fname, 'rb')
                    m = pickle.load(f)
                    f.close()
                    self.model_locks.release(tablename, modelid)
                    return m
                if type(modelid) == list:
                    models = {}
                    for i in modelid:
                        if not utils.is_int(i):
                            raise utils.BayesDBError('Invalid modelid: %s' % str(modelid))
                        models[i] = get_single_model(int(i))
                    return models
                elif utils.is_int(modelid):
                    return get_single_model(int(modelid))
                else:
                    raise utils.BayesDBError('Invalid modelid: %s' % str(modelid))
            else:
                # Return all the models
                models = {}
                self.model_locks.acquire_table(tablename)
                fnames = os.listdir(models_dir)
                for fname in fnames:
                    if fname.startswith('model_'):
                        model_id = fname[6:]  # remove preceding 'model_'
                        model_id = int(model_id[:-4])  # remove trailing '.pkl' and cast to int
                        full_fname = os.path.join(models_dir, fname)
                        f = open(full_fname, 'rb')
                        m = pickle.load(f)
                        f.close()
                        models[model_id] = m
                self.model_locks.release_table(tablename)
                return models
        else:
            # Backwards compatibility with old model style.
            self.model_locks.acquire_table(tablename)
            try:
                f = open(os.path.join(self.data_dir, tablename, 'models.pkl'), 'rb')
                models = pickle.load(f)
                f.close()
                if modelid is not None:
                    ret = models[modelid]
                else:
                    ret = models
                self.model_locks.release_table(tablename)
                return ret
            except IOError:
                self.model_locks.release_table(tablename)
                return {}

    def get_column_labels(self, tablename):
        try:
            f = open(os.path.join(self.data_dir, tablename, 'column_labels.pkl'), 'rb')
            column_labels = pickle.load(f)
            f.close()
            return column_labels
        except IOError:
            return dict()

    def get_column_lists(self, tablename):
        try:
            f = open(os.path.join(self.data_dir, tablename, 'column_lists.pkl'), 'rb')
            column_lists = pickle.load(f)
            f.close()
            return column_lists
        except IOError:
            return dict()

    def get_row_lists(self, tablename):
        try:
            f = open(os.path.join(self.data_dir, tablename, 'row_lists.pkl'), 'rb')
            row_lists = pickle.load(f)
            f.close()
            return row_lists
        except IOError:
            return dict()

    def column_list_exists(self, tablename, list_name):
        '''
        Return True if list_name or list_name_0 exist as column lists.
        '''
        column_lists = self.get_column_lists(tablename)
        result = list_name in column_lists or '%s_0' % list_name in column_lists
        return result

    def row_list_exists(self, tablename, list_name):
        '''
        Return True if list_name or list_name_0 exist as column lists.
        '''
        row_lists = self.get_row_lists(tablename)
        result = list_name in row_lists or '%s_0' % list_name in row_lists
        return result

    def get_user_metadata(self, tablename):
        try:
            f = open(os.path.join(self.data_dir, tablename, 'user_metadata.pkl'), 'rb')
            column_labels = pickle.load(f)
            f.close()
            return column_labels
        except IOError:
            return dict()

    def add_user_metadata(self, tablename, metadata_key, metadata_value):
        user_metadata = self.get_user_metadata(tablename)
        user_metadata[metadata_key.lower()] = metadata_value
        self.write_user_metadata(tablename, user_metadata)

    def add_column_label(self, tablename, column_name, column_label):
        column_labels = self.get_column_labels(tablename)
        column_labels[column_name.lower()] = column_label
        self.write_column_labels(tablename, column_labels)

    def add_column_list(self, tablename, column_list_name, column_list):
        column_lists = self.get_column_lists(tablename)
        column_lists[column_list_name] = column_list
        self.write_column_lists(tablename, column_lists)

    def add_row_list(self, tablename, row_list_name, row_list):
        row_lists = self.get_row_lists(tablename)
        row_lists[row_list_name] = row_list
        self.write_row_lists(tablename, row_lists)

    def get_column_label(self, tablename, column_name):
        column_labels = self.get_column_labels(tablename)
        if column_name.lower() in column_labels:
            return column_labels[column_name.lower()]
        else:
            raise utils.BayesDBError('Column %s in btable %s has no label.'
                                     % (column_name, tablename))

    def get_column_list(self, tablename, column_list):
        column_lists = self.get_column_lists(tablename)
        if column_list in column_lists:
            return column_lists[column_list]
        else:
            raise utils.BayesDBColumnListDoesNotExistError(column_list, tablename)

    def get_row_list(self, tablename, row_list):
        row_lists = self.get_row_lists(tablename)
        if row_list in row_lists:
            return row_lists[row_list]
        else:
            raise utils.BayesDBRowListDoesNotExistError(row_list, tablename)

    def write_model(self, tablename, model, modelid):
        # Make models dir
        models_dir = os.path.join(self.data_dir, tablename, 'models')
        if not os.path.exists(models_dir):
            os.makedirs(models_dir)

        self.model_locks.acquire(tablename, modelid)
        model_f = open(os.path.join(models_dir, 'model_%d.pkl' % modelid), 'wb')
        pickle.dump(model, model_f, pickle.HIGHEST_PROTOCOL)
        model_f.close()
        self.model_locks.release(tablename, modelid)

    def write_models(self, tablename, models):
        # Make models dir
        models_dir = os.path.join(self.data_dir, tablename, 'models')
        if not os.path.exists(models_dir):
            os.makedirs(models_dir)

        # Write each model individually
        for i, v in models.items():
            self.model_locks.acquire(tablename, modelid)
            model_f = open(os.path.join(models_dir, 'model_%d.pkl' % i), 'wb')
            pickle.dump(v, model_f, pickle.HIGHEST_PROTOCOL)
            model_f.close()
            self.model_locks.release(tablename, modelid)

    def write_column_labels(self, tablename, column_labels):
        column_labels_f = open(os.path.join(self.data_dir, tablename, 'column_labels.pkl'), 'wb')
        pickle.dump(column_labels, column_labels_f, pickle.HIGHEST_PROTOCOL)
        column_labels_f.close()

    def write_user_metadata(self, tablename, user_metadata):
        user_metadata_f = open(os.path.join(self.data_dir, tablename, 'user_metadata.pkl'), 'wb')
        pickle.dump(user_metadata, user_metadata_f, pickle.HIGHEST_PROTOCOL)
        user_metadata_f.close()

    def write_column_lists(self, tablename, column_lists):
        column_lists_f = open(os.path.join(self.data_dir, tablename, 'column_lists.pkl'), 'wb')
        pickle.dump(column_lists, column_lists_f, pickle.HIGHEST_PROTOCOL)
        column_lists_f.close()

    def write_row_lists(self, tablename, row_lists):
        row_lists_f = open(os.path.join(self.data_dir, tablename, 'row_lists.pkl'), 'wb')
        pickle.dump(row_lists, row_lists_f, pickle.HIGHEST_PROTOCOL)
        row_lists_f.close()

    def drop_btable(self, tablename):
        """Delete a single btable."""
        if tablename in self.btable_index:
            path = os.path.join(self.data_dir, tablename)
            if os.path.isdir(path):
                shutil.rmtree(path)
            self.remove_btable_from_index(tablename)
        return 0

    def list_btables(self):
        """Return a list of all btable names."""
        return self.btable_index

    def drop_models(self, tablename, model_ids='all'):
        """ Delete a single model, or all, if model_ids == 'all' or None. """
        models_dir = os.path.join(self.data_dir, tablename, 'models')
        if os.path.exists(models_dir):
            if model_ids is None or model_ids == 'all':
                self.model_locks.acquire_table(tablename)
                fnames = os.listdir(models_dir)
                for fname in fnames:
                    if 'model_' in fname:
                        os.remove(os.path.join(models_dir, fname))
                self.model_locks.drop_all(tablename)
                self.model_locks.release_table(tablename)
            else:
                for modelid in model_ids:
                    self.model_locks.acquire(tablename, modelid)
                    fname = os.path.join(models_dir, 'model_%d.pkl' % modelid)
                    if os.path.exists(fname):
                        os.remove(fname)
                    self.model_locks.drop(tablename, modelid)
                    self.model_locks.release(tablename, modelid)
        else:
            # If models in old style, convert to new style, save, and retry.
            models = self.get_models(tablename)
            self.write_models(tablename, models)
            self.drop_models(tablename, model_ids)

    def get_latent_states(self, tablename, modelid=None):
        """Return X_L_list, X_D_list, and M_c"""
        metadata = self.get_metadata(tablename)
        models = self.get_models(tablename, modelid)
        if None in models.values():
            raise utils.BayesDBError('Invalid model id. Use "SHOW MODELS FOR <btable>" to see '
                                     'valid model ids.')
        M_c = metadata['M_c']
        X_L_list = [model['X_L'] for model in models.values()]
        X_D_list = [model['X_D'] for model in models.values()]
        return (X_L_list, X_D_list, M_c)

    def get_metadata_and_table(self, tablename):
        """Return M_c and M_r and T"""
        metadata = self.get_metadata(tablename)
        M_c = metadata['M_c']
        M_r = metadata['M_r']
        T = metadata['T']
        return M_c, M_r, T

    def get_metadata_and_table_full(self, tablename):
        """Return M_c_full and M_r_full and T_full"""
        metadata_full = self.get_metadata_full(tablename)
        M_c_full = metadata_full['M_c_full']
        M_r_full = metadata_full['M_r_full']
        T_full = metadata_full['T_full']
        return M_c_full, M_r_full, T_full

    def has_models(self, tablename):
        return self.get_max_model_id(tablename) != -1

    def get_max_model_id(self, tablename, models=None):
        """Get the highest model id, and -1 if there are no models.
        Model indexing starts at 0 when models exist."""
        model_ids = self.get_model_ids(tablename, models)
        if len(model_ids) == 0:
            return -1
        else:
            return max(model_ids)

    def get_model_ids(self, tablename, models=None):
        """Get the list of modelids for a table."""
        if models is not None:
            model_ids = models.keys()
        else:
            models_dir = os.path.join(self.data_dir, tablename, 'models')
            if not os.path.exists(models_dir):
                model_ids = []
            else:
                model_ids = []
                fnames = os.listdir(models_dir)
                for fname in fnames:
                    if fname.startswith('model_'):
                        model_id = fname[6:]  # remove preceding 'model_'
                        model_id = int(model_id[:-4])  # remove trailing '.pkl' and cast to int
                        model_ids.append(model_id)
        return model_ids

    def get_cctypes(self, tablename):
        """Access the table's current cctypes."""
        metadata = self.get_metadata(tablename)
        return metadata['cctypes']

    def get_cctypes_full(self, tablename):
        """Access the table's current cctypes."""
        metadata_full = self.get_metadata_full(tablename)
        return metadata_full['cctypes_full']

    def get_colnames(self, tablename):
        """
        Return column names of analysis columns
        """
        M_c = self.get_metadata(tablename)['M_c']
        colnames = utils.get_all_column_names_in_original_order(M_c)
        return colnames

    def get_colnames_full(self, tablename):
        """
        Return column names of all columns
        """
        M_c_full = self.get_metadata_full(tablename)['M_c_full']
        colnames = utils.get_all_column_names_in_original_order(M_c_full)
        return colnames

    def get_schema(self, tablename):
        """
        Return colnames and cctypes of analysis columns
        """
        colnames = self.get_colnames(tablename)
        cctypes = self.get_cctypes(tablename)
        M_c = self.get_metadata(tablename)['M_c']
        schema = dict()
        for colname, cctype in zip(colnames, cctypes):
            schema[colname] = dict()
            schema[colname]['cctype'] = cctype
            if cctype == 'cyclic':
                col_idx = M_c['name_to_idx'][colname]
                schema[colname]['parameters'] = M_c['column_metadata'][col_idx]['parameters']
            else:
                schema[colname]['parameters'] = None
        return schema

    def get_schema_full(self, tablename):
        """
        Return colnames and cctypes of all columns
        """
        colnames_full = self.get_colnames_full(tablename)
        cctypes_full = self.get_cctypes_full(tablename)
        schema_full = dict()
        for colname, cctype in zip(colnames_full, cctypes_full):
            schema_full[colname] = cctype
        return schema_full

    def get_key_column_name(self, tablename):
        metadata_full = self.get_metadata_full(tablename)
        try:
            key_column_index = metadata_full['cctypes_full'].index('key')
        except ValueError as e:
            print("If your btable was created with an older version of BayesDB, it might not have "
                  "a default key column.")
            key_column_index = metadata_full['cctypes_full'].index('key')
        key_column_name = metadata_full['M_c_full']['idx_to_name'][str(key_column_index)]
        return key_column_name

    def update_metadata(self, tablename, M_r=None, M_c=None, T=None, cctypes=None):
        """Overwrite M_r, M_c, and T (not cctypes) for the table."""
        metadata = self.get_metadata(tablename)
        if M_r:
            metadata['M_r'] = M_r
        if M_c:
            metadata['M_c'] = M_c
        if T:
            metadata['T'] = T
        if cctypes:
            metadata['cctypes'] = cctypes
        f = open(os.path.join(self.data_dir, tablename, 'metadata.pkl'), 'wb')
        pickle.dump(metadata, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    def update_metadata_full(self, tablename, M_r_full=None, M_c_full=None, T_full=None,
                             cctypes_full=None):
        """Overwrite M_r, M_c, and T (not cctypes) for the table."""
        metadata = self.get_metadata_full(tablename)
        if M_r_full:
            metadata['M_r_full'] = M_r_full
        if M_c_full:
            metadata['M_c_full'] = M_c_full
        if T_full:
            metadata['T_full'] = T_full
        if cctypes_full:
            metadata['cctypes_full'] = cctypes_full
        f = open(os.path.join(self.data_dir, tablename, 'metadata_full.pkl'), 'wb')
        pickle.dump(metadata, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    def gen_default_codebook(self, M_c_full):
        """ Generates default metadata.

        Args:
            M_c_full (dict): M_c_full metadata object

        Returns:
            list of dict: num_cols length codebook, where each entry is a dict with keys:
                short_name, description, and valuemap with defaults colname, 'No description', and
                None, repsectively.
        """
        if 'column_codebook' in M_c_full:
            empty_codebook_entires = False
            for entry in M_c_full['column_codebook']:
                if entry is None:
                    empty_codebook_entires = True
                    break
            if not empty_codebook_entires:
                return

        column_codebook = [None]*len(M_c_full['name_to_idx'])

        for colname, idx in M_c_full['name_to_idx'].iteritems():
            column_codebook[idx] = {
                'description': 'No description.',
                'short_name': colname,
                'value_map': None
            }

        return column_codebook

    def update_codebook(self, tablename, codebook):
        """ Adds a codebook to a btable from a .csv

        Args:
            tablename (str): Name of the btable
            codebook (list of dict): The processed codebook. Should come from client because we
                assume that .csv is stored on the client machine.
        """
        # metadata_full = self.get_metadata_full(tablename)
        # M_c_full = metadata_full['M_c_full']
        #
        # if isinstance(codebook, dict):
        #     name_to_idx = M_c_full['name_to_idx']
        #     column_codebook = [None]*len(name_to_idx)
        #     for colname, idx in name_to_idx.iteritems():
        #         if colname in codebook.keys():
        #             column_codebook[idx] = codebook[colname]
        #     M_c_full['column_codebook'] = column_codebook
        #
        # elif isinstance(codebook, list):
        #     if not isinstance(codebook[0], dict):
        #         raise TypeError("Error: Invalid codebook.")
        #     M_c_full['column_codebook'] = codebook
        #
        # self.update_metadata_full(tablename, M_c_full=M_c_full)

        # Bax: I'm killing this feature (maybe temporarily) because there would be problems with
        # model metadata if value maps are updated. If no models have been created and there is a
        # codebook in place, it is trival for the user to create a new btable.
        raise NotImplementedError('UPDATE CODEBOOK is not currently implemented.\n Updating value '
                                  'Please recreate your BTABLE')

    def add_default_codebook_to_metadata(self, tablename):
        """ Adds the default metadata to tablename"""
        metadata_full = self.get_metadata_full(tablename)
        column_codebook = self.gen_default_codebook(metadata_full['M_c_full'])
        self.update_codebook_from_csv(tablename, column_codebook)

    def check_if_table_exists(self, tablename):
        """Return true iff this tablename exists in the persistence layer."""
        return tablename in self.btable_index

    def update_schema(self, tablename, mappings):
        """
        mappings is a dict of column name to 'cyclic', 'numerical', 'categorical', 'ignore', or
        'key'.
        TODO: can we get rid of cctypes?
        """
        metadata_full = self.get_metadata_full(tablename)
        cctypes_full = metadata_full['cctypes_full']
        M_c_full = metadata_full['M_c_full']
        raw_T_full = metadata_full['raw_T_full']
        colnames_full = utils.get_all_column_names_in_original_order(M_c_full)
        try:
            parameters_full = [x['parameters'] for x in M_c_full['column_metadata']]
        except KeyError:
            print('WARNING: resetting parameters to defaults. Please check these values with '
                  'DESCRIBE and adjust them manually if necessary.')
            parameters_full = []
            for md in M_c_full['column_metadata']:
                if 'dirichlet' in md['modeltype']:
                    params = {
                        'cardinality': len(md['value_to_code'])
                    }
                elif'vonmises' in md['modeltype']:
                    params = {
                        'min': 0.0,
                        'max': 2.0*3.14159265358979323846264338328
                    }
                else:
                    params = None

                parameters_full.append(params)

            parameters_full = [None for _ in range(len(M_c_full['column_metadata']))]

        # Now, update cctypes_full (cctypes updated later, after removing ignores).
        mapping_set = 'numerical', 'categorical', 'ignore', 'key', 'cyclic'

        for colname, mapping in mappings.items():
            cctype = mapping['cctype']
            parameters = mapping['parameters']

            if colname.lower() not in M_c_full['name_to_idx']:
                raise utils.BayesDBError('Error: column %s does not exist.' % colname)
            elif cctype not in mapping_set:
                raise utils.BayesDBError('Error: datatype %s is not one of the valid datatypes: %s.'
                                         % (mapping, str(mapping_set)))

            cidx = M_c_full['name_to_idx'][colname.lower()]

            # If the column's current type is key, don't allow the change.
            if cctypes_full[cidx] == 'key':
                raise utils.BayesDBError('Error: %s is already set as the table key. To change its '
                                         'type, reload the table using CREATE BTABLE and choose a '
                                         'different key column.' % colname.lower())
            # If the user tries to change a column to key, it's easier to reload the table, since at
            # this point there aren't models anyways. Eventually we can build this in if it's
            # desirable.
            elif cctype == 'key':
                raise utils.BayesDBError('Error: key column already exists. To choose a different '
                                         'key, reload the table using CREATE BTABLE')

            cctypes_full[cidx] = cctype
            parameters_full[cidx] = parameters

        # Make sure there isn't more than one key.
        assert len(filter(lambda x: x == 'key', cctypes_full)) == 1

        T_full, M_r_full, M_c_full, _ = data_utils.gen_T_and_metadata(colnames_full, raw_T_full,
                                                                      cctypes=cctypes_full,
                                                                      parameters=parameters_full)

        # Variables without "_full" don't include ignored columns.
        raw_T, cctypes, colnames, parameters = data_utils.remove_ignore_cols(raw_T_full,
                                                                             cctypes_full,
                                                                             colnames_full,
                                                                             parameters_full)
        T, M_r, M_c, _ = data_utils.gen_T_and_metadata(colnames, raw_T, cctypes=cctypes,
                                                       parameters=parameters)

        # Now, put cctypes, T, M_c, and M_r back into the DB
        self.update_metadata(tablename, M_r, M_c, T, cctypes)
        self.update_metadata_full(tablename, M_r_full, M_c_full, T_full, cctypes_full)

        return self.get_metadata_full(tablename)

    def create_btable(self, tablename, cctypes_full, cctypes, T, M_r, M_c, T_full, M_r_full,
                      M_c_full, raw_T_full, T_sub=None):
        """
        This function is called to create a btable.
        It creates the table's persistence directory, saves data.csv and metadata.pkl.
        Creates models.pkl as empty dict.

        """
        # Make directory for table
        if not os.path.exists(os.path.join(self.data_dir, tablename)):
            os.mkdir(os.path.join(self.data_dir, tablename))

        # Write metadata and metadata_full
        metadata_full = dict(M_c_full=M_c_full, M_r_full=M_r_full, T_full=T_full,
                             cctypes_full=cctypes_full, raw_T_full=raw_T_full)
        self.write_metadata_full(tablename, metadata_full)
        metadata = dict(M_c=M_c, M_r=M_r, T=T, cctypes=cctypes, T_sub=T_sub)
        self.write_metadata(tablename, metadata)

        # Write models
        models = {}
        self.write_models(tablename, models)

        # Write column labels
        column_labels = dict()
        self.write_column_labels(tablename, column_labels)

        # Write column lists
        column_lists = dict()
        self.write_column_lists(tablename, column_lists)

        # Write row lists
        row_lists = dict()
        self.write_row_lists(tablename, row_lists)

        # Add to btable name index
        self.add_btable_to_index(tablename)

    def upgrade_btable(self, tablename, cctypes_full, T_full, M_r_full, M_c_full, raw_T_full):
        """
        This function is called to upgrade a btable.
        Right now, it only updates the _full version of variables, because those are the only
        possible upgraded attributes, because the only possible upgrade reasons are:
        1. Creation of metadata_full
        2. Addition of key column.

        This will leave models, column labels, columns lists, and row lists intact.
        """
        # Write metadata_full
        metadata_full = dict(M_c_full=M_c_full, M_r_full=M_r_full, T_full=T_full,
                             cctypes_full=cctypes_full, raw_T_full=raw_T_full)
        self.write_metadata_full(tablename, metadata_full)

    def add_models(self, tablename, model_list):
        """
        Add a set of models (X_Ls and X_Ds) to a table (the table does not need to
        already have models).

        parameter model_list is a list of dicts, where each dict contains the keys
        X_L, X_D, and iterations.
        """
        # Model indexing starts at 0 (and is -1 if none exist)
        max_model_id = self.get_max_model_id(tablename)
        for i, m in enumerate(model_list):
            modelid = max_model_id + 1 + i
            self.write_model(tablename, m, modelid)

    def update_models(self, tablename, modelids, X_L_list, X_D_list, diagnostics_data_list,
                      increment_iterations_list=None, increment_time_list=None):
        """
        TODO: UNUSED WITH NEW ANALYZE, DELETE

        Overwrite all models by id, and append diagnostic info.

        Args:
            tablename (type): Btable name.
            X_L_list (list): List of crosscat X_L metadata from intialize or analyze
            X_D (list): List of Crosscat X_D metadata from intialize or analyze
            diagnostics_data (list of dict): List of diagnostics dict. Keys vary depending on
                tests avalible indiagnostics_utils. Each key holds only a single value.
            modelids (list of in): List of model indices.
            increment_iterations_list (list int): Number of analyze iterations since last update
                for each model.
            increment_time_list (list of float): Time (seconds) spent analyzing since last update
                for each model.
        """

        for idx, modelid in enumerate(modelids):
            if increment_iterations_list is not None:
                increment_iterations = increment_iterations[idx]
                increment_time = increment_time[idx]
            else:
                increment_iterations = 0
                increment_time = 0
            self.update_model(tablename, X_L_list[idx], X_D_list[idx], diagnostics_data_list[idx],
                              modelid, increment_iterations, increment_time)

    def update_model(self, tablename, X_L, X_D, diagnostics_data, modelid,
                     increment_iterations=0, increment_time=0):
        """ Overwrite a certain model by id.

        Overwrites current metadata and updates, trials, time, and diagnostics (if applicable).

        Args:
            tablename (type): Btable name.
            X_L (dict): Crosscat X_L metadata from intialize or analyze
            X_D (list of list of int): Crosscat X_D metadata from intialize or analyze
            diagnostics_data (dict): Diagnostics dict. Keys vary depending on tests avalible in
                diagnostics_utils. Each key holds only a single value.
            modelid (int): Model index.
            increment_iterations (int): Number of analyze iterations since last update.
            increment_time (float): Time (seconds) spent analyzing since last update.
        """
        assert type(modelid) == int
        self.model_locks.acquire(tablename, modelid)
        model = self.get_models(tablename, modelid)

        model['X_L'] = X_L
        model['X_D'] = X_D
        model['time'] = model.get('time') or 0

        if increment_iterations > 0:  # If any iterations were performed
            model['iterations'] += increment_iterations
            model['time'] += increment_time

            diagnostics_data['time'] = model['time']
            diagnostics_data['iterations'] = model['iterations']

            if 'diagnostics' not in model.keys():
                model['diagnostics'] = dict()
                for key, value in diagnostics_data.iteritems():
                    model['diagnostics'][key] = [value]
                model['diagnostics']['iterations'] = [0]
                model['diagnostics']['time'] = [0.0]
            else:
                # don't append diagnostics if this is the same timepoint
                if model['diagnostics']['iterations'][-1] != diagnostics_data['iterations'] and \
                        model['diagnostics']['time'][-1] != diagnostics_data['time']:
                    for key, value in diagnostics_data.iteritems():
                        if key not in model['diagnostics'].keys():
                            model['diagnostics'][key] = [value]
                        else:
                            model['diagnostics'][key].append(value)

        self.write_model(tablename, model, modelid)
        self.model_locks.release(tablename, modelid)
